# Copyright 2018-2020 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for computing Autograd gradients of quantum functions.
"""

import pytest
import autograd
import pennylane.numpy as anp  # only to be used inside classical computational nodes
import numpy as np

import pennylane as qml
from gate_data import Rotx as Rx, Roty as Ry, Rotz as Rz


alpha = 0.5  # displacement in tests
hbar = 2
mag_alphas = np.linspace(0, 1.5, 5)
thetas = np.linspace(-2 * np.pi, 2 * np.pi, 8)
sqz_vals = np.linspace(0.0, 1.0, 5)

cv_ops = [getattr(qml.ops, name) for name in qml.ops._cv__ops__]
analytic_cv_ops = [cls for cls in cv_ops if cls.supports_parameter_shift]


class PolyN(qml.ops.PolyXP):
    "Mimics NumberOperator using the arbitrary 2nd order observable interface. Results should be identical."

    def __init__(self, wires):
        hbar = 2
        q = np.diag([-0.5, 0.5 / hbar, 0.5 / hbar])
        super().__init__(q, wires=wires)
        self.name = "PolyXP"


@pytest.fixture(scope="module")
def gaussian_dev():
    return qml.device("default.gaussian", wires=2)


@pytest.fixture(scope="module")
def grad_fn_R(gaussian_dev):
    @qml.qnode(gaussian_dev)
    def circuit(y):
        qml.Displacement(alpha, 0.0, wires=[0])
        qml.Rotation(y, wires=[0])
        return qml.expval(qml.X(0))

    return autograd.grad(circuit)


@pytest.fixture(scope="module")
def grad_fn_BS(gaussian_dev):
    @qml.qnode(gaussian_dev)
    def circuit(y):
        qml.Displacement(alpha, 0.0, wires=[0])
        qml.Beamsplitter(y, 0, wires=[0, 1])
        return qml.expval(qml.X(0))

    return autograd.grad(circuit)


@pytest.fixture(scope="module")
def grad_fn_D(gaussian_dev):
    @qml.qnode(gaussian_dev)
    def circuit(r, phi):
        qml.Displacement(r, phi, wires=[0])
        return qml.expval(qml.X(0))

    return autograd.grad(circuit)


@pytest.fixture(scope="module")
def grad_fn_S(gaussian_dev):
    @qml.qnode(gaussian_dev)
    def circuit(y):
        qml.Displacement(alpha, 0.0, wires=[0])
        qml.Squeezing(y, 0.0, wires=[0])
        return qml.expval(qml.X(0))

    return autograd.grad(circuit)


@pytest.fixture(scope="module")
def grad_fn_S_Fock(gaussian_dev):
    @qml.qnode(gaussian_dev)
    def circuit(y):
        qml.Squeezing(y, 0.0, wires=[0])
        return qml.expval(qml.FockStateProjector(np.array([2, 0]), wires=[0, 1]))

    return autograd.grad(circuit)


class TestCVGradient:
    """Tests of the automatic gradient method for CV circuits."""

    @pytest.mark.parametrize("theta", thetas)
    def test_rotation_gradient(self, theta, grad_fn_R, tol):
        "Tests that the automatic gradient of a phase space rotation is correct."

        autograd_val = grad_fn_R(theta)
        # qfunc evalutes to hbar * alpha * cos(theta)
        manualgrad_val = -hbar * alpha * np.sin(theta)
        assert autograd_val == pytest.approx(manualgrad_val, abs=tol)

    @pytest.mark.parametrize("theta", thetas)
    def test_beamsplitter_gradient(self, theta, grad_fn_BS, tol):
        "Tests that the automatic gradient of a beamsplitter is correct."

        autograd_val = grad_fn_BS(theta)
        # qfunc evalutes to hbar * alpha * cos(theta)
        manualgrad_val = -hbar * alpha * np.sin(theta)
        assert autograd_val == pytest.approx(manualgrad_val, abs=tol)

    @pytest.mark.parametrize("mag", mag_alphas)
    @pytest.mark.parametrize("theta", thetas)
    def test_displacement_gradient(self, mag, theta, grad_fn_D, tol):
        "Tests that the automatic gradient of a phase space displacement is correct."

        # alpha = mag * np.exp(1j * theta)
        autograd_val = grad_fn_D(mag, theta)
        # qfunc evalutes to hbar * Re(alpha)
        manualgrad_val = hbar * np.cos(theta)
        assert autograd_val == pytest.approx(manualgrad_val, abs=tol)

    @pytest.mark.parametrize("r", sqz_vals)
    def test_squeeze_gradient(self, r, grad_fn_S, tol):
        "Tests that the automatic gradient of a phase space squeezing is correct."

        autograd_val = grad_fn_S(r)
        # qfunc evaluates to -exp(-r) * hbar * Re(alpha)
        manualgrad_val = -np.exp(-r) * hbar * alpha
        assert autograd_val == pytest.approx(manualgrad_val, abs=tol)

    @pytest.mark.parametrize("r", sqz_vals[1:])  # formula is not valid for r=0
    def test_number_state_gradient(self, r, grad_fn_S_Fock, tol):
        "Tests that the automatic gradient of a squeezed state with number state expectation is correct."

        # (d/dr) |<2|S(r)>|^2 = 0.5 tanh(r)^3 (2 csch(r)^2 - 1) sech(r)
        autograd_val = grad_fn_S_Fock(r)
        manualgrad_val = 0.5 * np.tanh(r) ** 3 * (2 / (np.sinh(r) ** 2) - 1) / np.cosh(r)
        assert autograd_val == pytest.approx(manualgrad_val, abs=tol)

    @pytest.mark.parametrize("O", [qml.ops.X, qml.ops.NumberOperator, PolyN])
    @pytest.mark.parametrize("G", analytic_cv_ops)
    def test_cv_gradients_gaussian_circuit(self, G, O, gaussian_dev, tol):
        """Tests that the gradients of circuits of gaussian gates match between the finite difference and analytic methods."""

        tol = 1e-5
        par = 0.4

        def circuit(x):
            args = [0.3] * G.num_params
            args[0] = x
            qml.Displacement(0.5, 0, wires=0)
            G(*args, wires=range(G.num_wires))
            qml.Beamsplitter(1.3, -2.3, wires=[0, 1])
            qml.Displacement(-0.5, 0.1, wires=0)
            qml.Squeezing(0.5, -1.5, wires=0)
            qml.Rotation(-1.1, wires=0)
            return qml.expval(O(wires=0))

        q = qml.QNode(circuit, gaussian_dev)
        val = q(par)

        grad_F = qml.gradients.finite_diff(q)(par)
        grad_A2 = qml.gradients.param_shift_cv(q, dev=gaussian_dev, force_order2=True)(par)
        if O.ev_order == 1:
            grad_A = qml.gradients.param_shift_cv(q, dev=gaussian_dev)(par)
            # the different methods agree
            assert grad_A == pytest.approx(grad_F, abs=tol)

        # the different methods agree
        assert grad_A2 == pytest.approx(grad_F, abs=tol)

    def test_cv_gradients_multiple_gate_parameters(self, gaussian_dev, tol):
        "Tests that gates with multiple free parameters yield correct gradients."
        par = [0.4, -0.3, -0.7, 0.2]

        def qf(r0, phi0, r1, phi1):
            qml.Squeezing(r0, phi0, wires=[0])
            qml.Squeezing(r1, phi1, wires=[0])
            return qml.expval(qml.NumberOperator(0))

        q = qml.QNode(qf, gaussian_dev)
        q(*par)
        grad_F = qml.gradients.finite_diff(q)(*par)
        grad_A = qml.gradients.param_shift_cv(q, dev=gaussian_dev)(*par)
        grad_A2 = qml.gradients.param_shift_cv(q, dev=gaussian_dev, force_order2=True)(*par)

        # the different methods agree
        assert grad_A == pytest.approx(grad_F, abs=tol)
        assert grad_A2 == pytest.approx(grad_F, abs=tol)

        # check against the known analytic formula
        r0, phi0, r1, phi1 = par
        dn = np.zeros([4])
        dn[0] = np.cosh(2 * r1) * np.sinh(2 * r0) + np.cos(phi0 - phi1) * np.cosh(2 * r0) * np.sinh(
            2 * r1
        )
        dn[1] = -0.5 * np.sin(phi0 - phi1) * np.sinh(2 * r0) * np.sinh(2 * r1)
        dn[2] = np.cos(phi0 - phi1) * np.cosh(2 * r1) * np.sinh(2 * r0) + np.cosh(2 * r0) * np.sinh(
            2 * r1
        )
        dn[3] = 0.5 * np.sin(phi0 - phi1) * np.sinh(2 * r0) * np.sinh(2 * r1)

        assert dn[np.newaxis, :] == pytest.approx(grad_F, abs=tol)

    def test_cv_gradients_repeated_gate_parameters(self, gaussian_dev, tol):
        "Tests that repeated use of a free parameter in a multi-parameter gate yield correct gradients."
        par = [0.2, 0.3]

        def qf(x, y):
            qml.Displacement(x, 0, wires=[0])
            qml.Squeezing(y, -1.3 * y, wires=[0])
            return qml.expval(qml.X(0))

        q = qml.QNode(qf, gaussian_dev)
        q(*par)
        grad_F = qml.gradients.finite_diff(q)(*par)
        grad_A = qml.gradients.param_shift_cv(q, dev=gaussian_dev)(*par)
        grad_A2 = qml.gradients.param_shift_cv(q, dev=gaussian_dev, force_order2=True)(*par)

        # the different methods agree
        assert grad_A == pytest.approx(grad_F, abs=tol)
        assert grad_A2 == pytest.approx(grad_F, abs=tol)

    def test_cv_gradients_parameters_inside_array(self, gaussian_dev, tol):
        "Tests that free parameters inside an array passed to an Operation yield correct gradients."
        par = [0.4, 1.3]

        def qf(x, y):
            qml.Displacement(0.5, 0, wires=[0])
            qml.Squeezing(x, 0, wires=[0])
            M = np.zeros((5, 5))
            M[1, 1] = y
            M[1, 2] = 1.0
            M[2, 1] = 1.0
            return qml.expval(qml.PolyXP(M, [0, 1]))

        q = qml.QNode(qf, gaussian_dev)
        q(*par)
        grad_F = qml.gradients.finite_diff(q, hybrid=False)(*par)
        grad_A = qml.gradients.param_shift_cv(q, dev=gaussian_dev, hybrid=False)(*par)
        grad_A2 = qml.gradients.param_shift_cv(
            q, dev=gaussian_dev, force_order2=True, hybrid=False
        )(*par)

        # the different methods agree
        assert grad_A2 == pytest.approx(grad_F, abs=tol)

    def test_cv_gradient_fanout(self, gaussian_dev, tol):
        "Tests that qnodes can compute the correct gradient when the same parameter is used in multiple gates."
        par = [0.5, 1.3]

        def circuit(x, y):
            qml.Displacement(x, 0, wires=[0])
            qml.Rotation(y, wires=[0])
            qml.Displacement(0, x, wires=[0])
            return qml.expval(qml.X(0))

        q = qml.QNode(circuit, gaussian_dev)
        q(*par)
        grad_F = qml.gradients.finite_diff(q)(*par)
        grad_A = qml.gradients.param_shift_cv(q, dev=gaussian_dev)(*par)
        grad_A2 = qml.gradients.param_shift_cv(q, dev=gaussian_dev, force_order2=True)(*par)

        # the different methods agree
        assert grad_A == pytest.approx(grad_F, abs=tol)
        assert grad_A2 == pytest.approx(grad_F, abs=tol)

    @pytest.mark.parametrize("name", qml.ops._cv__ops__)
    def test_CVOperation_with_heisenberg_and_no_params(self, name, gaussian_dev, tol):
        """An integration test for CV gates that support analytic differentiation
        if succeeding the gate to be differentiated, but cannot be differentiated
        themselves (for example, they may be Gaussian but accept no parameters).

        This ensures that, assuming their _heisenberg_rep is defined, the quantum
        gradient analytic method can still be used, and returns the correct result.
        """

        cls = getattr(qml.ops, name)
        if cls.supports_heisenberg and (not cls.supports_parameter_shift):
            U = np.array(
                [
                    [0.51310276 + 0.81702166j, 0.13649626 + 0.22487759j],
                    [0.26300233 + 0.00556194j, -0.96414101 - 0.03508489j],
                ]
            )

            if cls.num_wires <= 0:
                w = list(range(2))
            else:
                w = list(range(cls.num_wires))

            def circuit(x):
                qml.Displacement(x, 0, wires=0)

                if cls.par_domain == "A":
                    cls(U, wires=w)
                else:
                    cls(wires=w)
                return qml.expval(qml.X(0))

            qnode = qml.QNode(circuit, gaussian_dev)
            qnode(0.5)
            grad_F = qml.gradients.finite_diff(qnode)(0.5)
            grad_A = qml.gradients.param_shift_cv(qnode, dev=gaussian_dev)(0.5)
            grad_A2 = qml.gradients.param_shift_cv(qnode, dev=gaussian_dev, force_order2=True)(0.5)

            # the different methods agree
            assert grad_A == pytest.approx(grad_F, abs=tol)
            assert grad_A2 == pytest.approx(grad_F, abs=tol)


class TestQubitGradient:
    """Tests of the automatic gradient method for qubit gates."""

    def test_RX_gradient(self, qubit_device_1_wire, tol):
        "Tests that the automatic gradient of a Pauli X-rotation is correct."

        @qml.qnode(qubit_device_1_wire)
        def circuit(x):
            qml.RX(x, wires=[0])
            return qml.expval(qml.PauliZ(0))

        grad_fn = autograd.grad(circuit)

        for theta in thetas:
            autograd_val = grad_fn(theta)
            manualgrad_val = (circuit(theta + np.pi / 2) - circuit(theta - np.pi / 2)) / 2
            assert autograd_val == pytest.approx(manualgrad_val, abs=tol)

    def test_RY_gradient(self, qubit_device_1_wire, tol):
        "Tests that the automatic gradient of a Pauli Y-rotation is correct."

        @qml.qnode(qubit_device_1_wire)
        def circuit(x):
            qml.RY(x, wires=[0])
            return qml.expval(qml.PauliZ(0))

        grad_fn = autograd.grad(circuit)

        for theta in thetas:
            autograd_val = grad_fn(theta)
            manualgrad_val = (circuit(theta + np.pi / 2) - circuit(theta - np.pi / 2)) / 2
            assert autograd_val == pytest.approx(manualgrad_val, abs=tol)

    def test_RZ_gradient(self, qubit_device_1_wire, tol):
        "Tests that the automatic gradient of a Pauli Z-rotation is correct."

        @qml.qnode(qubit_device_1_wire)
        def circuit(x):
            qml.RZ(x, wires=[0])
            return qml.expval(qml.PauliZ(0))

        grad_fn = autograd.grad(circuit)

        for theta in thetas:
            autograd_val = grad_fn(theta)
            manualgrad_val = (circuit(theta + np.pi / 2) - circuit(theta - np.pi / 2)) / 2
            assert autograd_val == pytest.approx(manualgrad_val, abs=tol)

    def test_Rot(self, qubit_device_1_wire, tol):
        "Tests that the automatic gradient of a arbitrary Euler-angle-parameterized gate is correct."

        @qml.qnode(qubit_device_1_wire)
        def circuit(x, y, z):
            qml.Rot(x, y, z, wires=[0])
            return qml.expval(qml.PauliZ(0))

        grad_fn = autograd.grad(circuit, argnum=[0, 1, 2])

        eye = np.eye(3)
        for theta in thetas:
            angle_inputs = np.array([theta, theta ** 3, np.sqrt(2) * theta])
            autograd_val = grad_fn(*angle_inputs)
            for idx in range(3):
                onehot_idx = eye[idx]
                param1 = angle_inputs + np.pi / 2 * onehot_idx
                param2 = angle_inputs - np.pi / 2 * onehot_idx
                manualgrad_val = (circuit(*param1) - circuit(*param2)) / 2
                assert autograd_val[idx] == pytest.approx(manualgrad_val, abs=tol)

    def test_U2(self, tol):
        """Tests that the gradient of an arbitrary U2 gate is correct"""
        dev = qml.device("default.qubit", wires=1)

        @qml.qnode(dev)
        def circuit(x, y):
            qml.QubitStateVector(1j * np.array([1, -1]) / np.sqrt(2), wires=[0])
            qml.U2(x, y, wires=[0])
            return qml.expval(qml.PauliX(0))

        phi = -0.234
        lam = 0.654

        res = circuit(phi, lam)
        expected = np.sin(lam) * np.sin(phi)
        assert np.allclose(res, expected, atol=tol, rtol=0)

        grad_fn = autograd.grad(circuit, argnum=[0, 1])
        res = grad_fn(phi, lam)
        expected = np.array([np.sin(lam) * np.cos(phi), np.cos(lam) * np.sin(phi)])
        assert np.allclose(res, expected, atol=tol, rtol=0)

    def test_U3(self, tol):
        """Tests that the gradient of an arbitrary U3 gate is correct"""
        dev = qml.device("default.qubit", wires=1)

        @qml.qnode(dev)
        def circuit(x, y, z):
            qml.QubitStateVector(1j * np.array([1, -1]) / np.sqrt(2), wires=[0])
            qml.U3(x, y, z, wires=[0])
            return qml.expval(qml.PauliX(0))

        theta = 0.543
        phi = -0.234
        lam = 0.654

        res = circuit(theta, phi, lam)
        expected = np.sin(lam) * np.sin(phi) - np.cos(theta) * np.cos(lam) * np.cos(phi)
        assert np.allclose(res, expected, atol=tol, rtol=0)

        grad_fn = autograd.grad(circuit, argnum=[0, 1, 2])
        res = grad_fn(theta, phi, lam)
        expected = np.array(
            [
                np.sin(theta) * np.cos(lam) * np.cos(phi),
                np.cos(theta) * np.cos(lam) * np.sin(phi) + np.sin(lam) * np.cos(phi),
                np.cos(theta) * np.sin(lam) * np.cos(phi) + np.cos(lam) * np.sin(phi),
            ]
        )
        assert np.allclose(res, expected, atol=tol, rtol=0)

    def test_qfunc_gradients(self, qubit_device_2_wires, tol):
        "Tests that the various ways of computing the gradient of a qfunc all agree."

        def circuit(x, y, z):
            qml.RX(x, wires=[0])
            qml.CNOT(wires=[0, 1])
            qml.RY(-1.6, wires=[0])
            qml.RY(y, wires=[1])
            qml.CNOT(wires=[1, 0])
            qml.RX(z, wires=[0])
            qml.CNOT(wires=[0, 1])
            return qml.expval(qml.PauliZ(0))

        qnode = qml.QNode(circuit, qubit_device_2_wires, diff_method="parameter-shift")
        params = np.array([0.1, -1.6, np.pi / 5])

        # manual gradients
        grad_fd1 = qml.gradients.finite_diff(qnode, approx_order=1)(*params)
        grad_fd2 = qml.gradients.finite_diff(qnode, approx_order=2)(*params)
        grad_angle = qml.gradients.param_shift(qnode)(*params)

        # automatic gradient
        grad_fn = qml.grad(qnode)
        grad_auto = grad_fn(*params)

        # gradients computed with different methods must agree
        assert grad_fd1 == pytest.approx(grad_fd2, abs=tol)
        assert grad_fd1 == pytest.approx(grad_angle, abs=tol)
        assert np.allclose(grad_fd1, grad_auto, atol=tol, rtol=0)

    def test_hybrid_gradients(self, qubit_device_2_wires, tol):
        "Tests that the various ways of computing the gradient of a hybrid computation all agree."

        # input data is the first parameter
        def classifier_circuit(in_data, x):
            qml.RX(in_data, wires=[0])
            qml.CNOT(wires=[0, 1])
            qml.RY(-1.6, wires=[0])
            qml.RY(in_data, wires=[1])
            qml.CNOT(wires=[1, 0])
            qml.RX(x, wires=[0])
            qml.CNOT(wires=[0, 1])
            return qml.expval(qml.PauliZ(0))

        classifier = qml.QNode(
            classifier_circuit, qubit_device_2_wires, diff_method="parameter-shift"
        )

        param = -0.1259
        in_data = qml.numpy.array([-0.1, -0.88, np.exp(0.5)], requires_grad=False)
        out_data = np.array([1.5, np.pi / 3, 0.0])

        def error(p):
            "Total square error of classifier predictions."
            ret = 0
            for d_in, d_out in zip(in_data, out_data):
                square_diff = (classifier(d_in, p) - d_out) ** 2
                ret = ret + square_diff
            return ret

        def d_error(p, grad_method):
            "Gradient of error, computed manually."
            ret = 0
            for d_in, d_out in zip(in_data, out_data):
                args = (d_in, p)
                diff = classifier(*args) - d_out
                ret = ret + 2 * diff * grad_method(classifier)(d_in, p)
            return ret

        y0 = error(param)
        grad = autograd.grad(error)
        grad_auto = grad(param)

        grad_fd1 = d_error(param, qml.gradients.finite_diff)
        grad_angle = d_error(param, qml.gradients.param_shift)

        # gradients computed with different methods must agree
        assert grad_fd1 == pytest.approx(grad_angle, abs=tol)
        assert grad_fd1 == pytest.approx(grad_auto, abs=tol)
        assert grad_angle == pytest.approx(grad_auto, abs=tol)

    def test_hybrid_gradients_autograd_numpy(self, qubit_device_2_wires, tol):
        "Test the gradient of a hybrid computation requiring autograd.numpy functions."

        def circuit(x, y):
            "Quantum node."
            qml.RX(x, wires=[0])
            qml.CNOT(wires=[0, 1])
            qml.RY(y, wires=[0])
            qml.CNOT(wires=[0, 1])
            return qml.expval(qml.PauliZ(0)), qml.expval(qml.PauliZ(1))

        quantum = qml.QNode(circuit, qubit_device_2_wires, diff_method="parameter-shift")

        def classical(p):
            "Classical node, requires autograd.numpy functions."
            return anp.exp(anp.sum(quantum(p[0], anp.log(p[1]))))

        def d_classical(a, b, grad_method):
            "Gradient of classical computed symbolically, can use normal numpy functions."
            val = classical((a, b))
            J = grad_method(quantum)(a, np.log(b))
            return val * np.array([J[0, 0] + J[1, 0], (J[0, 1] + J[1, 1]) / b])

        param = np.array([-0.1259, 1.53])
        y0 = classical(param)
        grad_classical = autograd.jacobian(classical)
        grad_auto = grad_classical(param)

        grad_fd1 = d_classical(*param, qml.gradients.finite_diff)
        grad_angle = d_classical(*param, qml.gradients.param_shift)

        # gradients computed with different methods must agree
        assert grad_fd1 == pytest.approx(grad_angle, abs=tol)
        assert grad_fd1 == pytest.approx(grad_auto, abs=tol)
        assert grad_angle == pytest.approx(grad_auto, abs=tol)

    def test_qnode_gradient_fanout(self, qubit_device_1_wire, tol):
        "Tests that the correct gradient is computed for qnodes which use the same parameter in multiple gates."

        def expZ(state):
            return np.abs(state[0]) ** 2 - np.abs(state[1]) ** 2

        extra_param = 0.31

        def circuit(reused_param, other_param):
            qml.RX(extra_param, wires=[0])
            qml.RY(reused_param, wires=[0])
            qml.RZ(other_param, wires=[0])
            qml.RX(reused_param, wires=[0])
            return qml.expval(qml.PauliZ(0))

        f = qml.QNode(circuit, qubit_device_1_wire)
        zero_state = np.array([1.0, 0.0])

        for reused_p in thetas:
            reused_p = reused_p ** 3 / 19
            for other_p in thetas:
                other_p = other_p ** 2 / 11

                # autograd gradient
                grad = autograd.grad(f)
                grad_eval = grad(reused_p, other_p)

                # manual gradient
                grad_true0 = (
                    expZ(
                        Rx(reused_p)
                        @ Rz(other_p)
                        @ Ry(reused_p + np.pi / 2)
                        @ Rx(extra_param)
                        @ zero_state
                    )
                    - expZ(
                        Rx(reused_p)
                        @ Rz(other_p)
                        @ Ry(reused_p - np.pi / 2)
                        @ Rx(extra_param)
                        @ zero_state
                    )
                ) / 2
                grad_true1 = (
                    expZ(
                        Rx(reused_p + np.pi / 2)
                        @ Rz(other_p)
                        @ Ry(reused_p)
                        @ Rx(extra_param)
                        @ zero_state
                    )
                    - expZ(
                        Rx(reused_p - np.pi / 2)
                        @ Rz(other_p)
                        @ Ry(reused_p)
                        @ Rx(extra_param)
                        @ zero_state
                    )
                ) / 2
                grad_true = grad_true0 + grad_true1  # product rule

                assert grad_eval == pytest.approx(grad_true, abs=tol)

    def test_gradient_exception_on_sample(self):
        """Tests that the proper exception is raised if differentiation of sampling is attempted."""
        dev = qml.device("default.qubit", wires=2, shots=1000)

        @qml.qnode(dev, diff_method="parameter-shift")
        def circuit(x):
            qml.RX(x, wires=[0])
            return qml.sample(qml.PauliZ(0)), qml.sample(qml.PauliX(1))

        with pytest.raises(
            qml.QuantumFunctionError,
            match="Circuits that include sampling can not be differentiated.",
        ):
            grad_fn = autograd.jacobian(circuit)
            grad_fn(1.0)

    def test_autograd_positional_non_trainable_warns_grad(self):
        """Test that a warning is raised if a positional argument without the
        requires_grad attribute set is passed when differentiating a QNode with a
        scalar output."""
        dev = qml.device("default.qubit", wires=5)

        @qml.qnode(dev)
        def test(x):
            qml.RY(x, wires=[0])
            return qml.expval(qml.PauliZ(0))

        with pytest.warns(
            UserWarning, match="inputs have to explicitly specify requires_grad=True"
        ):
            qml.grad(test)(0.3)

    def test_autograd_positional_non_trainable_warns_jacobian(self):
        """Test that a warning is raised if a positional argument without the
        requires_grad attribute set is passed when differentiating a QNode with a
        vector output."""
        dev = qml.device("default.qubit", wires=5)

        @qml.qnode(dev)
        def test(x):
            qml.RY(x, wires=[0])
            return qml.probs(wires=[0])

        with pytest.warns(
            UserWarning, match="inputs have to explicitly specify requires_grad=True"
        ):
            qml.jacobian(test)(0.3)

    def test_autograd_trainable_no_warn_grad(self, recwarn):
        """Test that no warning is raised if positional arguments are marked as
        trainable using the requires_grad attribute."""
        dev = qml.device("default.qubit", wires=5)

        @qml.qnode(dev)
        def test(x):
            qml.RZ(x, wires=[0])
            return qml.expval(qml.PauliZ(0))

        par = anp.array(0.3, requires_grad=True)
        qml.grad(test)(par)
        assert len(recwarn) == 0

    def test_autograd_trainable_argnum_no_warn_grad(self, recwarn):
        """Test that no warning is raised if positional arguments are marked as
        trainable using the argnum argument."""
        dev = qml.device("default.qubit", wires=5)

        @qml.qnode(dev)
        def test(x):
            qml.RZ(x, wires=[0])
            return qml.expval(qml.PauliZ(0))

        par = np.array(0.3)
        qml.grad(test, argnum=0)(par)
        assert len(recwarn) == 0

    def test_autograd_trainable_no_warn_jacobian(self, recwarn):
        """Test that no warning is raised if positional arguments are marked as
        trainable using the requires_grad attribute."""
        dev = qml.device("default.qubit", wires=5)

        @qml.qnode(dev)
        def test(x):
            qml.RZ(x, wires=[0])
            return qml.probs(wires=[0])

        par = anp.array(0.3, requires_grad=True)
        qml.jacobian(test)(par)
        assert len(recwarn) == 0

    def test_autograd_trainable_argnum_no_warn_jacobian(self, recwarn):
        """Test that no warning is raised if positional arguments are marked as
        trainable using the argnum argument."""
        dev = qml.device("default.qubit", wires=5)

        @qml.qnode(dev)
        def test(x):
            qml.RZ(x, wires=[0])
            return qml.probs(wires=[0])

        par = np.array(0.3)
        qml.jacobian(test, argnum=0)(par)
        assert len(recwarn) == 0


class TestFourTermParameterShifts:
    """Tests for quantum gradients that require a 4-term shift formula"""

    @pytest.mark.parametrize("G", [qml.CRX, qml.CRY, qml.CRZ])
    def test_controlled_rotation_gradient(self, G, tol):
        """Test gradient of controlled RX gate"""
        dev = qml.device("default.qubit", wires=2)
        b = 0.123

        @qml.qnode(dev, diff_method="parameter-shift")
        def circuit(b):
            qml.QubitStateVector(np.array([1.0, -1.0]) / np.sqrt(2), wires=0)
            G(b, wires=[0, 1])
            return qml.expval(qml.PauliX(0))

        res = circuit(b)
        assert np.allclose(res, -np.cos(b / 2), atol=tol, rtol=0)

        grad = qml.grad(circuit)(b)
        expected = np.sin(b / 2) / 2
        assert np.allclose(grad, expected, atol=tol, rtol=0)

    @pytest.mark.parametrize("theta", np.linspace(-2 * np.pi, np.pi, 7))
    def test_CRot_gradient(self, theta, tol):
        """Tests that the automatic gradient of a arbitrary controlled Euler-angle-parameterized
        gate is correct."""
        dev = qml.device("default.qubit", wires=2)
        a, b, c = np.array([theta, theta ** 3, np.sqrt(2) * theta])

        @qml.qnode(dev, diff_method="parameter-shift")
        def circuit(a, b, c):
            qml.QubitStateVector(np.array([1.0, -1.0]) / np.sqrt(2), wires=0)
            qml.CRot(a, b, c, wires=[0, 1])
            return qml.expval(qml.PauliX(0))

        res = circuit(a, b, c)
        expected = -np.cos(b / 2) * np.cos(0.5 * (a + c))
        assert np.allclose(res, expected, atol=tol, rtol=0)

        grad = qml.grad(circuit)(a, b, c)
        expected = np.array(
            [
                [
                    0.5 * np.cos(b / 2) * np.sin(0.5 * (a + c)),
                    0.5 * np.sin(b / 2) * np.cos(0.5 * (a + c)),
                    0.5 * np.cos(b / 2) * np.sin(0.5 * (a + c)),
                ]
            ]
        )
        assert np.allclose(grad, expected, atol=tol, rtol=0)



# Copyright 2018-2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# pylint: disable=protected-access
"""
Contains a utility class ``BooleanFn`` that allows logical composition
of functions with boolean output.
"""
import functools


class BooleanFn:
    r"""Wrapper for simple callables with boolean output that can be
    manipulated and combined with bit-wise operators.

    Args:
        fn (callable): Function to be wrapped. It must accept a single
            argument, and must return a boolean.

    **Example**

    Consider functions that filter numbers to lie in a certain domain.
    We may wrap them using ``BooleanFn``:

    >>> bigger_than_4 = qml.BooleanFn(lambda x: x > 4)
    >>> smaller_than_10 = qml.BooleanFn(lambda x: x < 10)
    >>> is_int = qml.BooleanFn(lambda x: isinstance(x, int))
    >>> bigger_than_4(5.2)
    True
    >>> smaller_than_10(20.1)
    False
    >>> is_int(2.3)
    False

    These can then be combined into a single callable using boolean operators,
    such as ``&``, logical and:

    >>> between_4_and_10 = bigger_than_4 & smaller_than_10
    >>> between_4_and_10(-3.2)
    False
    >>> between_4_and_10(9.9)
    True
    >>> between_4_and_10(19.7)
    False

    Other supported operators are ``|``, logical or, and ``~``, logical not:

    >>> smaller_equal_than_4 = ~bigger_than_4
    >>> smaller_than_10_or_int = smaller_than_10 | is_int

    .. warning::

        Note that Python conditional expressions are evaluated from left to right.
        As a result, the order of composition may matter, even though logical
        operators such as ``|`` and ``&`` are symmetric.

        For example:

        >>> is_int = qml.BooleanFn(lambda x: isinstance(x, int))
        >>> has_bit_length_3 = qml.BooleanFn(lambda x: x.bit_length()==3)
        >>> (is_int & has_bit_length_3)(4)
        True
        >>> (is_int & has_bit_length_3)(2.3)
        False
        >>> (has_bit_length_3 & is_int)(2.3)
        AttributeError: 'float' object has no attribute 'bit_length'

    """

    def __init__(self, fn):
        self.fn = fn
        functools.update_wrapper(self, fn)

    def __and__(self, other):
        return BooleanFn(lambda obj: self.fn(obj) and other.fn(obj))

    def __or__(self, other):
        return BooleanFn(lambda obj: self.fn(obj) or other.fn(obj))

    def __invert__(self):
        return BooleanFn(lambda obj: not self.fn(obj))

    def __call__(self, obj):
        return self.fn(obj)



# Copyright 2018-2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Contains the classical Jacobian transform.
"""
# pylint: disable=import-outside-toplevel
import pennylane as qml
from pennylane import numpy as np


def classical_jacobian(qnode, argnum=None, expand_fn=None):
    r"""Returns a function to extract the Jacobian
    matrix of the classical part of a QNode.

    This transform allows the classical dependence between the QNode
    arguments and the quantum gate arguments to be extracted.

    Args:
        qnode (pennylane.QNode): QNode to compute the (classical) Jacobian of
        argnum (int or Sequence[int]): indices of QNode arguments with respect to which
            the (classical) Jacobian is computed
        expand_fn (None or function): an expansion function (if required) to be applied to the
            QNode quantum tape before the classical Jacobian is computed

    Returns:
        function: Function which accepts the same arguments as the QNode.
        When called, this function will return the Jacobian of the QNode
        gate arguments with respect to the QNode arguments indexed by ``argnum``.

    **Example**

    Consider the following QNode:

    >>> @qml.qnode(dev)
    ... def circuit(weights):
    ...     qml.RX(weights[0], wires=0)
    ...     qml.RY(0.2 * weights[0], wires=1)
    ...     qml.RY(2.5, wires=0)
    ...     qml.RZ(weights[1] ** 2, wires=1)
    ...     qml.RX(weights[2], wires=1)
    ...     return qml.expval(qml.PauliZ(0) @ qml.PauliZ(1))

    We can use this transform to extract the relationship :math:`f: \mathbb{R}^n \rightarrow
    \mathbb{R}^m` between the input QNode arguments :math:`w` and the gate arguments :math:`g`, for
    a given value of the QNode arguments:

    >>> cjac_fn = qml.transforms.classical_jacobian(circuit)
    >>> weights = np.array([1., 1., 0.6], requires_grad=True)
    >>> cjac = cjac_fn(weights)
    >>> print(cjac)
    [[1.  0.  0. ]
     [0.2 0.  0. ]
     [0.  0.  0. ]
     [0.  1.2 0. ]
     [0.  0.  1. ]]

    The returned Jacobian has rows corresponding to gate arguments, and columns
    corresponding to QNode arguments; that is,

    .. math:: J_{ij} = \frac{\partial}{\partial g_i} f(w_j).

    We can see that:

    - The zeroth element of ``weights`` is repeated on the first two gates generated by the QNode.

    - The third row consisting of all zeros indicates that the third gate ``RY(2.5)`` does not
      depend on the ``weights``.

    - The quadratic dependence of the fourth gate argument yields :math:`2\cdot 0.6=1.2`.

    .. note::

        The QNode is constructed during this operation.

    For a QNode with multiple QNode arguments, the arguments with respect to which the
    Jacobian is computed can be controlled with the ``argnum`` keyword argument.
    The output for ``argnum=None`` depends on the backend:

    .. list-table:: Output format of ``classical_jacobian``
       :widths: 25 25 25 25
       :header-rows: 1

       * - Interface
         - ``argnum=None``
         - ``type(argnum)=int``
         - ``argnum=Sequence[int]``
       * - ``'autograd'``
         - ``tuple(arrays)`` [1]
         - ``array``
         - ``tuple(array)``
       * - ``'jax'``
         - ``array``
         - ``array``
         - ``tuple(array)``
       * - ``'tf'``
         - ``tuple(arrays)``
         - ``array``
         - ``tuple(array)``
       * - ``'torch'``
         - ``tuple(arrays)``
         - ``array``
         - ``tuple(array)``

    [1] If all QNode argument are scalars, the tuple is unpacked and the one-dimensional Jacobian
    arrays are stacked into one ``array``. If there only is one QNode argument, the tuple is
    unpacked as well. Both is due to the behaviour of ``qml.jacobian``.

    **Example with ``argnum``**

    >>> @qml.qnode(dev)
    ... def circuit(x, y, z):
    ...     qml.RX(qml.math.sin(x), wires=0)
    ...     qml.CNOT(wires=[0, 1])
    ...     qml.RY(y ** 2, wires=1)
    ...     qml.RZ(1 / z, wires=1)
    ...     return qml.expval(qml.PauliZ(0) @ qml.PauliZ(1))
    >>> jac_fn = qml.transforms.classical_jacobian(circuit, argnum=[1, 2])
    >>> x, y, z = np.array([0.1, -2.5, 0.71])
    >>> jac_fn(x, y, z)
    (array([-0., -5., -0.]), array([-0.        , -0.        , -1.98373339]))

    Only the Jacobians with respect to the arguments ``x`` and ``y`` were computed, and
    returned as a tuple of ``arrays``.

    """

    def classical_preprocessing(*args, **kwargs):
        """Returns the trainable gate parameters for a given QNode input."""
        trainable_only = kwargs.pop("_trainable_only", True)
        kwargs.pop("shots", None)
        qnode.construct(args, kwargs)
        tape = qnode.qtape

        if expand_fn is not None:
            tape = expand_fn(tape)

        return qml.math.stack(tape.get_parameters(trainable_only=trainable_only))

    if qnode.interface == "autograd":

        def _jacobian(*args, **kwargs):
            if argnum is None:
                jac = qml.jacobian(classical_preprocessing)(*args, **kwargs)
            elif np.isscalar(argnum):
                jac = qml.jacobian(classical_preprocessing, argnum=argnum)(*args, **kwargs)
            else:
                jac = tuple(
                    (
                        qml.jacobian(classical_preprocessing, argnum=i)(*args, **kwargs)
                        for i in argnum
                    )
                )
            return jac

        return _jacobian

    if qnode.interface == "torch":
        import torch

        def _jacobian(*args, **kwargs):  # pylint: disable=unused-argument
            jac = torch.autograd.functional.jacobian(classical_preprocessing, args)
            if argnum is not None:
                if np.isscalar(argnum):
                    jac = jac[argnum]
                else:
                    jac = tuple((jac[idx] for idx in argnum))
            return jac

        return _jacobian

    if qnode.interface == "jax":
        import jax

        argnum = 0 if argnum is None else argnum

        def _jacobian(*args, **kwargs):
            kwargs["_trainable_only"] = False
            return jax.jacobian(classical_preprocessing, argnums=argnum)(*args, **kwargs)

        return _jacobian

    if qnode.interface == "tf":
        import tensorflow as tf

        def _jacobian(*args, **kwargs):
            if np.isscalar(argnum):
                sub_args = args[argnum]
            elif argnum is None:
                sub_args = args
            else:
                sub_args = tuple((args[i] for i in argnum))

            with tf.GradientTape() as tape:
                gate_params = classical_preprocessing(*args, **kwargs)

            jac = tape.jacobian(gate_params, sub_args)
            return jac

        return _jacobian



"""Convenience gate representations for testing"""
import math
import cmath
import numpy as np

# ========================================================
#  fixed gates
# ========================================================

I = np.eye(2)

# Pauli matrices
X = np.array([[0, 1], [1, 0]])  #: Pauli-X matrix
Y = np.array([[0, -1j], [1j, 0]])  #: Pauli-Y matrix
Z = np.array([[1, 0], [0, -1]])  #: Pauli-Z matrix

H = np.array([[1, 1], [1, -1]]) / math.sqrt(2)  #: Hadamard gate

II = np.eye(4, dtype=np.complex128)
XX = np.array(np.kron(X, X), dtype=np.complex128)
YY = np.array(np.kron(Y, Y), dtype=np.complex128)

# Single-qubit projectors
StateZeroProjector = np.array([[1, 0], [0, 0]])
StateOneProjector = np.array([[0, 0], [0, 1]])

# Two qubit gates
CNOT = np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]])  #: CNOT gate
SWAP = np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]])  #: SWAP gate
ISWAP = np.array([[1, 0, 0, 0], [0, 0, 1j, 0], [0, 1j, 0, 0], [0, 0, 0, 1]])  #: ISWAP gate
SISWAP = np.array(
    [
        [1, 0, 0, 0],
        [0, 1 / math.sqrt(2), 1 / math.sqrt(2) * 1j, 0],
        [0, 1 / math.sqrt(2) * 1j, 1 / math.sqrt(2), 0],
        [0, 0, 0, 1],
    ]
)
CZ = np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, -1]])  #: CZ gate
S = np.array([[1, 0], [0, 1j]])  #: Phase Gate
T = np.array([[1, 0], [0, cmath.exp(1j * np.pi / 4)]])  #: T Gate
# Three qubit gates
CSWAP = np.array(
    [
        [1, 0, 0, 0, 0, 0, 0, 0],
        [0, 1, 0, 0, 0, 0, 0, 0],
        [0, 0, 1, 0, 0, 0, 0, 0],
        [0, 0, 0, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, 1, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 1, 0],
        [0, 0, 0, 0, 0, 1, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 1],
    ]
)  #: CSWAP gate

Toffoli = np.diag([1 for i in range(8)])
Toffoli[6:8, 6:8] = np.array([[0, 1], [1, 0]])

w = np.exp(2 * np.pi * 1j / 8)
QFT = (
    np.array(
        [
            [1, 1, 1, 1, 1, 1, 1, 1],
            [1, w, w ** 2, w ** 3, w ** 4, w ** 5, w ** 6, w ** 7],
            [1, w ** 2, w ** 4, w ** 6, 1, w ** 2, w ** 4, w ** 6],
            [1, w ** 3, w ** 6, w, w ** 4, w ** 7, w ** 2, w ** 5],
            [1, w ** 4, 1, w ** 4, 1, w ** 4, 1, w ** 4],
            [1, w ** 5, w ** 2, w ** 7, w ** 4, w, w ** 6, w ** 3],
            [1, w ** 6, w ** 4, w ** 2, 1, w ** 6, w ** 4, w ** 2],
            [1, w ** 7, w ** 6, w ** 5, w ** 4, w ** 3, w ** 2, w],
        ]
    )
    / np.sqrt(8)
)

# ========================================================
#  parametrized gates
# ========================================================


def Rphi(phi):
    r"""One-qubit phase shift.

    Args:
        phi (float): phase shift angle
    Returns:
        array: unitary 2x2 phase shift matrix
    """
    return np.array([[1, 0], [0, cmath.exp(1j * phi)]])


def Rotx(theta):
    r"""One-qubit rotation about the x axis.

    Args:
        theta (float): rotation angle
    Returns:
        array: unitary 2x2 rotation matrix :math:`e^{-i \sigma_x \theta/2}`
    """
    return math.cos(theta / 2) * I + 1j * math.sin(-theta / 2) * X


def Roty(theta):
    r"""One-qubit rotation about the y axis.

    Args:
        theta (float): rotation angle
    Returns:
        array: unitary 2x2 rotation matrix :math:`e^{-i \sigma_y \theta/2}`
    """
    return math.cos(theta / 2) * I + 1j * math.sin(-theta / 2) * Y


def Rotz(theta):
    r"""One-qubit rotation about the z axis.

    Args:
        theta (float): rotation angle
    Returns:
        array: unitary 2x2 rotation matrix :math:`e^{-i \sigma_z \theta/2}`
    """
    return math.cos(theta / 2) * I + 1j * math.sin(-theta / 2) * Z


def Rot3(a, b, c):
    r"""Arbitrary one-qubit rotation using three Euler angles.

    Args:
        a,b,c (float): rotation angles
    Returns:
        array: unitary 2x2 rotation matrix ``rz(c) @ ry(b) @ rz(a)``
    """
    return Rotz(c) @ (Roty(b) @ Rotz(a))


def CRotx(theta):
    r"""Two-qubit controlled rotation about the x axis.

    Args:
        theta (float): rotation angle
    Returns:
        array: unitary 4x4 rotation matrix :math:`|0\rangle\langle 0|\otimes \mathbb{I}+|1\rangle\langle 1|\otimes R_x(\theta)`
    """
    return np.array(
        [
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, math.cos(theta / 2), -1j * math.sin(theta / 2)],
            [0, 0, -1j * math.sin(theta / 2), math.cos(theta / 2)],
        ]
    )


def CRoty(theta):
    r"""Two-qubit controlled rotation about the y axis.

    Args:
        theta (float): rotation angle
    Returns:
        array: unitary 4x4 rotation matrix :math:`|0\rangle\langle 0|\otimes \mathbb{I}+|1\rangle\langle 1|\otimes R_y(\theta)`
    """
    return np.array(
        [
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, math.cos(theta / 2), -math.sin(theta / 2)],
            [0, 0, math.sin(theta / 2), math.cos(theta / 2)],
        ]
    )


def CRotz(theta):
    r"""Two-qubit controlled rotation about the z axis.

    Args:
        theta (float): rotation angle
    Returns:
        array: unitary 4x4 rotation matrix :math:`|0\rangle\langle 0|\otimes \mathbb{I}+|1\rangle\langle 1|\otimes R_z(\theta)`
    """
    return np.array(
        [
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, cmath.exp(-1j * theta / 2), 0],
            [0, 0, 0, cmath.exp(1j * theta / 2)],
        ]
    )


def CRot3(a, b, c):
    r"""Arbitrary two-qubit controlled rotation using three Euler angles.

    Args:
        a,b,c (float): rotation angles
    Returns:
        array: unitary 4x4 rotation matrix :math:`|0\rangle\langle 0|\otimes \mathbb{I}+|1\rangle\langle 1|\otimes R(a,b,c)`
    """
    return np.array(
        [
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [
                0,
                0,
                cmath.exp(-1j * (a + c) / 2) * math.cos(b / 2),
                -cmath.exp(1j * (a - c) / 2) * math.sin(b / 2),
            ],
            [
                0,
                0,
                cmath.exp(-1j * (a - c) / 2) * math.sin(b / 2),
                cmath.exp(1j * (a + c) / 2) * math.cos(b / 2),
            ],
        ]
    )


def MultiRZ1(theta):
    r"""Arbitrary multi Z rotation on one wire.

    Args:
        theta (float): rotation angle

    Returns:
        array: the one-wire MultiRZ matrix
    """
    return np.array([[np.exp(-1j * theta / 2), 0.0 + 0.0j], [0.0 + 0.0j, np.exp(1j * theta / 2)]])


def MultiRZ2(theta):
    r"""Arbitrary multi Z rotation on two wires.

    Args:
        theta (float): rotation angle

    Returns:
        array: the two-wire MultiRZ matrix
    """
    return np.array(
        [
            [np.exp(-1j * theta / 2), 0.0 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j],
            [0.0 + 0.0j, np.exp(1j * theta / 2), 0.0 + 0.0j, 0.0 + 0.0j],
            [0.0 + 0.0j, 0.0 + 0.0j, np.exp(1j * theta / 2), 0.0 + 0.0j],
            [0.0 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j, np.exp(-1j * theta / 2)],
        ]
    )


def IsingXX(phi):
    r"""Ising XX coupling gate

    .. math:: XX(\phi) = \begin{bmatrix}
        \cos(\phi / 2) & 0 & 0 & -i \sin(\phi / 2) \\
        0 & \cos(\phi / 2) & -i \sin(\phi / 2) & 0 \\
        0 & -i \sin(\phi / 2) & \cos(\phi / 2) & 0 \\
        -i \sin(\phi / 2) & 0 & 0 & \cos(\phi / 2)
        \end{bmatrix}.

    Args:
        phi (float): rotation angle :math:`\phi`
    Returns:
        array[complex]: unitary 4x4 rotation matrix
    """
    return np.cos(phi / 2) * II - 1j * np.sin(phi / 2) * XX


def IsingYY(phi):
    r"""Ising YY coupling gate.

    .. math:: YY(\phi) = \begin{bmatrix}
        \cos(\phi / 2) & 0 & 0 & i \sin(\phi / 2) \\
        0 & \cos(\phi / 2) & -i \sin(\phi / 2) & 0 \\
        0 & -i \sin(\phi / 2) & \cos(\phi / 2) & 0 \\
        i \sin(\phi / 2) & 0 & 0 & \cos(\phi / 2)
        \end{bmatrix}.

    Args:
        phi (float): rotation angle :math:`\phi`
    Returns:
        array[complex]: unitary 4x4 rotation matrix
    """
    return np.cos(phi / 2) * II - 1j * np.sin(phi / 2) * YY


def IsingZZ(phi):
    r"""Ising ZZ coupling gate

    .. math:: ZZ(\phi) = \begin{bmatrix}
        e^{-i \phi / 2} & 0 & 0 & 0 \\
        0 & e^{i \phi / 2} & 0 & 0 \\
        0 & 0 & e^{i \phi / 2} & 0 \\
        0 & 0 & 0 & e^{-i \phi / 2}
        \end{bmatrix}.

    Args:
        phi (float): rotation angle :math:`\phi`
    Returns:
        array[complex]: unitary 4x4 rotation matrix
    """
    e_m = np.exp(-1j * phi / 2)
    e = np.exp(1j * phi / 2)
    return np.array([[e_m, 0, 0, 0], [0, e, 0, 0], [0, 0, e, 0], [0, 0, 0, e_m]])


def ControlledPhaseShift(phi):
    r"""Controlled phase shift.

    Args:
        phi (float): rotation angle

    Returns:
        array: the two-wire controlled-phase matrix
    """
    return np.diag([1, 1, 1, np.exp(1j * phi)])


def SingleExcitation(phi):
    r"""Single excitation rotation.

    Args:
        phi (float): rotation angle

    Returns:
        array: the two-qubit Givens rotation describing the single excitation operation
    """

    return np.array(
        [
            [1, 0, 0, 0],
            [0, np.cos(phi / 2), -np.sin(phi / 2), 0],
            [0, np.sin(phi / 2), np.cos(phi / 2), 0],
            [0, 0, 0, 1],
        ]
    )


def SingleExcitationPlus(phi):
    r"""Single excitation rotation with positive phase shift.

    Args:
    phi (float): rotation angle

    Returns:
        array: the two-qubit Givens rotation describing the single excitation operation
    """

    return np.array(
        [
            [np.exp(1j * phi / 2), 0, 0, 0],
            [0, np.cos(phi / 2), -np.sin(phi / 2), 0],
            [0, np.sin(phi / 2), np.cos(phi / 2), 0],
            [0, 0, 0, np.exp(1j * phi / 2)],
        ]
    )


def SingleExcitationMinus(phi):
    r"""Single excitation rotation with negative phase shift.

    Args:
        phi (float): rotation angle

    Returns:
        array: the two-qubit matrix describing the operation
    """

    return np.array(
        [
            [np.exp(-1j * phi / 2), 0, 0, 0],
            [0, np.cos(phi / 2), -np.sin(phi / 2), 0],
            [0, np.sin(phi / 2), np.cos(phi / 2), 0],
            [0, 0, 0, np.exp(-1j * phi / 2)],
        ]
    )


def DoubleExcitation(phi):
    r"""Double excitation rotation.

    Args:
        phi (float): rotation angle
    Returns:
        array: the four-qubit Givens rotation describing the double excitation
    """

    c = math.cos(phi / 2)
    s = math.sin(phi / 2)

    U = np.eye(16)
    U[3, 3] = c  # 3 (dec) = 0011 (bin)
    U[3, 12] = -s  # 12 (dec) = 1100 (bin)
    U[12, 3] = s
    U[12, 12] = c

    return U


def DoubleExcitationPlus(phi):
    r"""Double excitation rotation with positive phase shift.

    Args:
        phi (float): rotation angle

    Returns:
        array: the four-qubit matrix describing the operation
    """

    c = math.cos(phi / 2)
    s = math.sin(phi / 2)
    e = cmath.exp(1j * phi / 2)

    U = e * np.eye(16, dtype=np.complex64)
    U[3, 3] = c  # 3 (dec) = 0011 (bin)
    U[3, 12] = -s  # 12 (dec) = 1100 (bin)
    U[12, 3] = s
    U[12, 12] = c

    return U


def DoubleExcitationMinus(phi):
    r"""Double excitation rotation with negative phase shift.

    Args:
        phi (float): rotation angle
    Returns:
        array: the four-qubit matrix describing the operation
    """

    c = math.cos(phi / 2)
    s = math.sin(phi / 2)
    e = cmath.exp(-1j * phi / 2)

    U = e * np.eye(16, dtype=np.complex64)
    U[3, 3] = c  # 3 (dec) = 0011 (bin)
    U[3, 12] = -s  # 12 (dec) = 1100 (bin)
    U[12, 3] = s
    U[12, 12] = c

    return U


def OrbitalRotation(phi):
    r"""Quantum number preserving four-qubit one-parameter gate.

    Args:
        phi (float): rotation angle
    Returns:
        array: the four-qubit matrix describing the operation
    """

    c = math.cos(phi / 2)
    s = math.sin(phi / 2)

    return np.array(
        [
            [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, c, 0, 0, -s, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, c, 0, 0, 0, 0, 0, -s, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, c ** 2, 0, 0, -c * s, 0, 0, -c * s, 0, 0, s ** 2, 0, 0, 0],
            [0, s, 0, 0, c, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, c * s, 0, 0, c ** 2, 0, 0, -(s ** 2), 0, 0, -c * s, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, c, 0, 0, 0, 0, 0, -s, 0, 0],
            [0, 0, s, 0, 0, 0, 0, 0, c, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, c * s, 0, 0, -(s ** 2), 0, 0, c ** 2, 0, 0, -c * s, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, c, 0, 0, -s, 0],
            [0, 0, 0, s ** 2, 0, 0, c * s, 0, 0, c * s, 0, 0, c ** 2, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, s, 0, 0, 0, 0, 0, c, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, s, 0, 0, c, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
        ]
    )



# Copyright 2018-2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module contains the functions needed for performing the self-consistent-field calculations.
"""

import itertools

import autograd.numpy as anp
from pennylane.hf.matrices import (
    generate_core_matrix,
    generate_overlap_matrix,
    generate_repulsion_tensor,
    molecular_density_matrix,
)


def generate_scf(mol, n_steps=50, tol=1e-8):
    r"""Return a function that performs the self-consistent-field calculations.

    In the Hartree-Fock method, molecular orbitals are typically constructed as a linear combination
    of atomic orbitals

    .. math::

        \phi_i(r) = \sum_{\mu} C_{\mu i} \chi_{\mu}(r),

    with coefficients :math:`C_{\mu i}` that are initially unknown. The self-consistent-field
    iterations are performed to find a converged set of molecular orbital coefficients that minimize
    the total energy of the molecular system. This optimization problem can be reduced to solving a
    linear system of equations which are usually written as

    .. math::

        FC = SCE,

    where :math:`E` is a diagonal matrix of eigenvalues, representing the molecular orbital
    energies, :math:`C` is the matrix of molecular orbital coefficients, :math:`S` is the overlap
    matrix and :math:`F` is the Fock matrix, which also depends on the coefficients. Fixing an
    initial guess :math:`C_0`, the corresponding :math:`F_0` is built and the system
    :math:`F_0C_0 = SC_0E` is solved to obtain a solution :math:`C_1`. This process is iteratively
    repeated until the coefficients are converged.

    The key step in in this process is constructing the Fock matrix which is defined as

    .. math::

        F = H + \frac{1}{2} J - K,

    where :math:`H`, :math:`J` and :math:`K` are the core Hamiltonian matrix, Coulomb matrix and
    exchange matrix, respectively. The entries of :math:`H` are computed from the electronic kinetic
    energy and the electron-nuclear attraction integrals, which are integrals over atomic basis
    functions. The elements of the :math:`J` and :math:`K` matrices are obtained from the Coulomb
    and exchange integrals over the basis functions.

    Following the procedure in
    [`Lehtola et al. Molecules 2020, 25, 1218 <https://www.mdpi.com/1420-3049/25/5/1218>`_], we
    express the molecular orbital coefficients in terms of a matrix :math:`X` as
    :math:`C = X \tilde{C}` which gives the following transformed equation

    .. math::

         \tilde{F} \tilde{C} = \tilde{S} \tilde{C} E,

    where :math:`\tilde{F} = X^T F X`, :math:`\tilde{S} = X^T S X` and :math:`S` is the overlap
    matrix. We chose :math:`X` such that :math:`\tilde{S} = 1` as

    .. math::

        X = V \Lambda^{-1/2} V^T,

    where :math:`V` and :math:`\Lambda` are the eigenvectors and eigenvalues of :math:`S`,
    respectively. This gives the eigenvalue equation

    .. math::

         \tilde{F}\tilde{C} = \tilde{C}E,

    which is solved with conventional methods iteratively.

    Args:
        mol (Molecule): the molecule object
        n_steps (int): the number of iterations
        tol (float): convergence tolerance

    Returns:
        function: function that performs the self-consistent-field calculations

    **Example**

    >>> symbols  = ['H', 'H']
    >>> geometry = np.array([[0.0, 0.0, 0.0], [0.0, 0.0, 1.0]], requires_grad = False)
    >>> alpha = np.array([[3.42525091, 0.62391373, 0.1688554],
    >>>                   [3.42525091, 0.62391373, 0.1688554]], requires_grad=True)
    >>> mol = qml.hf.Molecule(symbols, geometry, alpha=alpha)
    >>> args = [alpha]
    >>> v_fock, coeffs, fock_matrix, h_core, repulsion_tensor = generate_scf(mol)(*args)
    >>> v_fock
    array([-0.67578019,  0.94181155])
    """

    def scf(*args):
        r"""Perform the self-consistent-field iterations.

        Args:
            args (array[array[float]]): initial values of the differentiable parameters

        Returns:
            tuple(array[float]): eigenvalues of the Fock matrix, molecular orbital coefficients,
            Fock matrix, core matrix
        """
        basis_functions = mol.basis_set
        charges = mol.nuclear_charges
        r = mol.coordinates
        n_electron = mol.n_electrons

        if r.requires_grad:
            args_r = [[args[0][i]] * mol.n_basis[i] for i in range(len(mol.n_basis))]
            args_ = [*args] + [anp.vstack(list(itertools.chain(*args_r)))]
            repulsion_tensor = generate_repulsion_tensor(basis_functions)(*args_[1:])
            s = generate_overlap_matrix(basis_functions)(*args_[1:])
            h_core = generate_core_matrix(basis_functions, charges, r)(*args_)
        else:
            repulsion_tensor = generate_repulsion_tensor(basis_functions)(*args)
            s = generate_overlap_matrix(basis_functions)(*args)
            h_core = generate_core_matrix(basis_functions, charges, r)(*args)

        s = s + anp.diag(anp.random.rand(len(s)) * 1.0e-12)

        w, v = anp.linalg.eigh(s)
        x = v @ anp.diag(anp.array([1 / anp.sqrt(i) for i in w])) @ v.T

        eigvals, w_fock = anp.linalg.eigh(x.T @ h_core @ x)  # initial guess for the scf problem
        coeffs = x @ w_fock

        p = molecular_density_matrix(n_electron, coeffs)

        for _ in range(n_steps):

            j = anp.einsum("pqrs,rs->pq", repulsion_tensor, p)
            k = anp.einsum("psqr,rs->pq", repulsion_tensor, p)

            fock_matrix = h_core + 2 * j - k

            eigvals, w_fock = anp.linalg.eigh(x.T @ fock_matrix @ x)

            coeffs = x @ w_fock

            p_update = molecular_density_matrix(n_electron, coeffs)

            if anp.linalg.norm(p_update - p) <= tol:
                break

            p = p_update

        return eigvals, coeffs, fock_matrix, h_core, repulsion_tensor

    return scf


def nuclear_energy(charges, r):
    r"""Return a function that computes the nuclear-repulsion energy.

    The nuclear-repulsion energy is computed as

    .. math::

        \sum_{i>j}^n \frac{q_i q_j}{r_{ij}},

    where :math:`q`, :math:`r` and :math:`n` denote the nuclear charges (atomic numbers), nuclear
    positions and the number of nuclei, respectively.

    Args:
        charges (list[int]): nuclear charges in atomic units
        r (array[float]): nuclear positions

    Returns:
        function: function that computes the nuclear-repulsion energy

    **Example**

    >>> symbols  = ['H', 'F']
    >>> geometry = np.array([[0.0, 0.0, 0.0], [0.0, 0.0, 2.0]], requires_grad = True)
    >>> mol = qml.hf.Molecule(symbols, geometry)
    >>> args = [mol.coordinates]
    >>> e = nuclear_energy(mol.nuclear_charges, mol.coordinates)(*args)
    >>> print(e)
    4.5
    """

    def nuclear(*args):
        r"""Compute the nuclear-repulsion energy.

        Args:
            args (array[array[float]]): initial values of the differentiable parameters

        Returns:
            array[float]: nuclear-repulsion energy
        """
        if r.requires_grad:
            coor = args[0]
        else:
            coor = r
        e = 0
        for i, r1 in enumerate(coor):
            for j, r2 in enumerate(coor[i + 1 :]):
                e = e + (charges[i] * charges[i + j + 1] / anp.linalg.norm(r1 - r2))
        return e

    return nuclear


def hf_energy(mol):
    r"""Return a function that computes the Hartree-Fock energy.

    Args:
        mol (Molecule): the molecule object

    Returns:
        function: function that computes the Hartree-Fock energy

    **Example**

    >>> symbols  = ['H', 'H']
    >>> geometry = np.array([[0.0, 0.0, 0.0], [0.0, 0.0, 1.0]], requires_grad = False)
    >>> alpha = np.array([[3.42525091, 0.62391373, 0.1688554],
    >>>                   [3.42525091, 0.62391373, 0.1688554]], requires_grad=True)
    >>> mol = qml.hf.Molecule(symbols, geometry, alpha=alpha)
    >>> args = [alpha]
    >>> hf_energy(mol)(*args)
    -1.065999461545263
    """

    def energy(*args):
        r"""Compute the Hartree-Fock energy.

        Args:
            args (array[array[float]]): initial values of the differentiable parameters

        Returns:
            float: the Hartree-Fock energy
        """
        _, coeffs, fock_matrix, h_core, _ = generate_scf(mol)(*args)
        e_rep = nuclear_energy(mol.nuclear_charges, mol.coordinates)(*args)
        e_elec = anp.einsum(
            "pq,qp", fock_matrix + h_core, molecular_density_matrix(mol.n_electrons, coeffs)
        )
        return e_elec + e_rep

    return energy



# Copyright 2018-2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Transform for merging adjacent rotations of the same type in a quantum circuit."""

from pennylane import apply
from pennylane.transforms import qfunc_transform
from pennylane.math import allclose, stack, cast_like, zeros

from pennylane.ops.qubit.attributes import composable_rotations
from .optimization_utils import find_next_gate, fuse_rot_angles


@qfunc_transform
def merge_rotations(tape, atol=1e-8, include_gates=None):
    r"""Quantum function transform to combine rotation gates of the same type
    that act sequentially.

    If the combination of two rotation produces an angle that is close to 0,
    neither gate will be applied.

    Args:
        qfunc (function): A quantum function.
        atol (float): After fusion of gates, if the fused angle :math:`\theta` is such that
            :math:`|\theta|\leq \text{atol}`, no rotation gate will be applied.
        include_gates (None or list[str]): A list of specific operations to merge. If
            set to ``None`` (default), all operations in the
            `~.pennylane.ops.qubit.attributes.composable_rotations` attribute will be merged. Otherwise,
            only the operations whose names match those in the list will undergo merging.

    Returns:
        function: the transformed quantum function

    **Example**

    Consider the following quantum function.

    .. code-block:: python

        def qfunc(x, y, z):
            qml.RX(x, wires=0)
            qml.RX(y, wires=0)
            qml.CNOT(wires=[1, 2])
            qml.RY(y, wires=1)
            qml.Hadamard(wires=2)
            qml.CRZ(z, wires=[2, 0])
            qml.RY(-y, wires=1)
            return qml.expval(qml.PauliZ(0))

    The circuit before optimization:

    >>> dev = qml.device('default.qubit', wires=3)
    >>> qnode = qml.QNode(qfunc, dev)
    >>> print(qml.draw(qnode)(1, 2, 3))
     0: ───RX(1)──RX(2)──────────╭RZ(3)──┤ ⟨Z⟩
     1: ──╭C──────RY(2)──RY(-2)──│───────┤
     2: ──╰X──────H──────────────╰C──────┤

    By inspection, we can combine the two ``RX`` rotations on the first qubit.
    On the second qubit, we have a cumulative angle of 0, and the gates will cancel.

    >>> optimized_qfunc = merge_rotations()(qfunc)
    >>> optimized_qnode = qml.QNode(optimized_qfunc, dev)
    >>> print(qml.draw(optimized_qnode)(1, 2, 3))
     0: ───RX(3)─────╭RZ(3)──┤ ⟨Z⟩
     1: ──╭C─────────│───────┤
     2: ──╰X──────H──╰C──────┤

    It is also possible to explicitly specify which rotations ``merge_rotations`` should
    be merged using the ``include_gates`` argument. For example, if in the above
    circuit we wanted only to merge the "RX" gates, we could do so as follows:

    >>> optimized_qfunc = merge_rotations(include_gates=["RX"])(qfunc)
    >>> optimized_qnode = qml.QNode(optimized_qfunc, dev)
    >>> print(qml.draw(optimized_qnode)(1, 2, 3))
     0: ───RX(3)─────────────────╭RZ(3)──┤ ⟨Z⟩
     1: ──╭C──────RY(2)──RY(-2)──│───────┤
     2: ──╰X──────H──────────────╰C──────┤
    """
    # Make a working copy of the list to traverse
    list_copy = tape.operations.copy()

    while len(list_copy) > 0:
        current_gate = list_copy[0]

        # If a specific list of operations is specified, check and see if our
        # op is in it, then try to merge. If not, queue and move on.
        if include_gates is not None:
            if current_gate.name not in include_gates:
                apply(current_gate)
                list_copy.pop(0)
                continue

        # Check if the rotation is composable; if it is not, move on.
        if not current_gate in composable_rotations:
            apply(current_gate)
            list_copy.pop(0)
            continue

        # Find the next gate that acts on the same wires
        next_gate_idx = find_next_gate(current_gate.wires, list_copy[1:])

        # If no such gate is found (either there simply is none, or there are other gates
        # "in the way", queue the operation and move on
        if next_gate_idx is None:
            apply(current_gate)
            list_copy.pop(0)
            continue

        # We need to use stack to get this to work and be differentiable in all interfaces
        cumulative_angles = stack(current_gate.parameters)

        # As long as there is a valid next gate, check if we can merge the angles
        while next_gate_idx is not None:
            # Get the next gate
            next_gate = list_copy[next_gate_idx + 1]

            # If next gate is of the same type, we can merge the angles
            if current_gate.name == next_gate.name and current_gate.wires == next_gate.wires:
                list_copy.pop(next_gate_idx + 1)

                # The Rot gate must be treated separately
                if current_gate.name == "Rot":
                    cumulative_angles = fuse_rot_angles(
                        cumulative_angles, cast_like(stack(next_gate.parameters), cumulative_angles)
                    )
                # Other, single-parameter rotation gates just have the angle summed
                else:
                    cumulative_angles = cumulative_angles + cast_like(
                        stack(next_gate.parameters), cumulative_angles
                    )
            # If it is not, we need to stop
            else:
                break

            # If we did merge, look now at the next gate
            next_gate_idx = find_next_gate(current_gate.wires, list_copy[1:])

        if not allclose(cumulative_angles, zeros(len(cumulative_angles)), atol=atol, rtol=0):
            current_gate.__class__(*cumulative_angles, wires=current_gate.wires)

        # Remove the first gate gate from the working list
        list_copy.pop(0)

    # Queue the measurements normally
    for m in tape.measurements:
        apply(m)



# Copyright 2018-2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""
Layers are trainable templates that are typically repeated, using different adjustable parameters in each repetition.
They implement a transformation from a quantum state to another quantum state.
"""

from .strongly_entangling import StronglyEntanglingLayers
from .random import RandomLayers
from .cv_neural_net import CVNeuralNetLayers
from .simplified_two_design import SimplifiedTwoDesign
from .basic_entangler import BasicEntanglerLayers
from .particle_conserving_u2 import ParticleConservingU2
from .particle_conserving_u1 import ParticleConservingU1
from .gate_fabric import GateFabric



# Copyright 2018-2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This file contains functionalities that simplify working with kernels.
"""
from pennylane import numpy as np


def square_kernel_matrix(X, kernel, assume_normalized_kernel=False):
    r"""Computes the square matrix of pairwise kernel values for a given dataset.

    Args:
        X (list[datapoint]): List of datapoints
        kernel ((datapoint, datapoint) -> float): Kernel function that maps
            datapoints to kernel value.
        assume_normalized_kernel (bool, optional): Assume that the kernel is normalized, in
            which case the diagonal of the kernel matrix is set to 1, avoiding unnecessary
            computations.

    Returns:
        array[float]: The square matrix of kernel values.

    **Example:**

    Consider a simple kernel function based on :class:`~.templates.embeddings.AngleEmbedding`:

    .. code-block :: python

        dev = qml.device('default.qubit', wires=2, shots=None)
        @qml.qnode(dev)
        def circuit(x1, x2):
            qml.templates.AngleEmbedding(x1, wires=dev.wires)
            qml.adjoint(qml.templates.AngleEmbedding)(x2, wires=dev.wires)
            return qml.probs(wires=dev.wires)

        kernel = lambda x1, x2: circuit(x1, x2)[0]

    We can then compute the kernel matrix on a set of 4 (random) feature
    vectors ``X`` via

    >>> X = np.random.random((4, 2))
    >>> qml.kernels.square_kernel_matrix(X, kernel)
    tensor([[1.        , 0.9532702 , 0.96864001, 0.90932897],
            [0.9532702 , 1.        , 0.99727485, 0.95685561],
            [0.96864001, 0.99727485, 1.        , 0.96605621],
            [0.90932897, 0.95685561, 0.96605621, 1.        ]], requires_grad=True)
    """
    N = len(X)
    matrix = [0] * N ** 2

    for i in range(N):
        for j in range(i, N):
            if assume_normalized_kernel and i == j:
                matrix[N * i + j] = 1.0
            else:
                matrix[N * i + j] = kernel(X[i], X[j])
                matrix[N * j + i] = matrix[N * i + j]

    return np.array(matrix).reshape((N, N))


def kernel_matrix(X1, X2, kernel):
    r"""Computes the matrix of pairwise kernel values for two given datasets.

    Args:
        X1 (list[datapoint]): List of datapoints (first argument)
        X2 (list[datapoint]): List of datapoints (second argument)
        kernel ((datapoint, datapoint) -> float): Kernel function that maps datapoints to kernel value.

    Returns:
        array[float]: The square matrix of kernel values.

    **Example:**

    Consider a simple kernel function based on :class:`~.templates.embeddings.AngleEmbedding`:

    .. code-block :: python

        dev = qml.device('default.qubit', wires=2, shots=None)
        @qml.qnode(dev)
        def circuit(x1, x2):
            qml.templates.AngleEmbedding(x1, wires=dev.wires)
            qml.adjoint(qml.templates.AngleEmbedding)(x2, wires=dev.wires)
            return qml.probs(wires=dev.wires)

        kernel = lambda x1, x2: circuit(x1, x2)[0]

    With this method we can systematically evaluate the kernel function ``kernel`` on
    pairs of datapoints, where the points stem from different datasets, like a training
    and a test dataset.

    >>> X_train = np.random.random((4,2))
    >>> X_test = np.random.random((3,2))
    >>> qml.kernels.kernel_matrix(X_train, X_test, kernel)
    tensor([[0.88875298, 0.90655175, 0.89926447],
            [0.93762197, 0.98163781, 0.93076383],
            [0.91977339, 0.9799841 , 0.91582698],
            [0.80376818, 0.98720925, 0.79349212]], requires_grad=True)

    As we can see, for :math:`n` and :math:`m` datapoints in the first and second
    dataset respectively, the output matrix has the shape :math:`n\times m`.
    """
    N = len(X1)
    M = len(X2)

    matrix = [0] * N * M
    for i in range(N):
        for j in range(M):
            matrix[M * i + j] = kernel(X1[i], X2[j])

    return np.array(matrix).reshape((N, M))



# Copyright 2018-2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Contains a transform that computes the frequency spectrum of a quantum
circuit including classical preprocessing within the QNode."""
from collections import OrderedDict
from functools import wraps
from inspect import signature
from itertools import product

import numpy as np
import pennylane as qml

from .utils import get_spectrum, join_spectra


def _process_ids(encoding_args, argnum, qnode):
    r"""Process the passed ``encoding_args`` and ``argnum`` or infer them from
    the QNode signature.

    Args:
        encoding_args (dict[str, list[tuple]] or set): Parameter index dictionary;
            keys are argument names, values are index tuples for that argument
            or an ``Ellipsis``. If a ``set``, all values are set to ``Ellipsis``
        argnum (list[int]): Numerical indices for arguments
        qnode (QNode): QNode to infer the ``encoding_args`` and ``argnum`` from
            if both are ``None``
    Returns:
        OrderedDict[str, list[tuple]]: Ordered parameter index dictionary;
            keys are argument names, values are index tuples for that argument
            or an ``Ellipsis``
        list[int]: Numerical indices for arguments

    In ``qnode_spectrum`` both ``encoding_args`` and ``argnum`` are required.
    However, they can be inferred from one another and even from the QNode signature,
    which is done in this helper function, using the following rules/design choices:

      - If ``argnum`` is provided, the QNode arguments with the indices in ``argnum``
        are considered and added to ``encoding_args`` with an ``Ellipsis``, meaning
        that for array-valued arguments all parameters are considered in
        ``qnode_spectrum``.
      - If ``encoding_args`` is provided and is a dictionary, it is preserved
        up to arguments that do not appear in the QNode. Also, it is converted to
        an ``OrderedDict``, inferring the ordering from the QNode arguments.
        Passing a set with ``keys`` instead is an alias for
        ``{key: ... for key in keys}``.
        ``argnum`` will contain the indices of these arguments.
      - If both ``encoding_args`` and ``argnum`` are passed, ``encoding_args`` takes
        precedence over ``argnum``, in particular ``argnum`` is overwritten.
      - If neither is passed, all arguments of the passed QNode that do not have a
        default value defined are considered
        and their value is an ``Ellipsis``, so that all parameters of array-valued
        arguments will be considered in ``qnode_spectrum``.

    **Example**

    As an example, consider the qnode

    >>> @qml.qnode(dev)
    >>> def circuit(a, b, c, x=2):
    ...     return qml.expval(qml.PauliX(0))

    which takes arguments:

    >>> a = np.array([2.4, 1.2, 3.1])
    >>> b = 0.2
    >>> c = np.arange(20, dtype=float).reshape((2, 5, 2))

    Then we may use the following inputs

    >>> encoding_args = {"a": [(1,), (2,)], "c": ..., "x": [()]}
    >>> argnum = [2, 0]

    in various combinations:

    >>> _process_ids(encoding_args, None, circuit)
    (OrderedDict([('a', [(1,), (2,)]), ('c', Ellipsis), ('x', [()])]), [0, 2, 3])

    The first output, ``encoding_args``, essentially is unchanged, it simply was ordered in
    the order of the QNode arguments. The second output, ``argnum``, contains all three
    argument indices because all of ``a``, ``b``, and ``c`` appear in ``encoding_args``.
    If we in addition pass ``argnum``, it is ignored:

    >>> _process_ids(encoding_args, argnum, circuit)
    (OrderedDict([('a', [(1,), (2,)]), ('c', Ellipsis), ('x', [()])]), [0, 2, 3])

    Only if we leave out ``encoding_args`` does it make a difference:

    >>> _process_ids(None, argnum, circuit)
    (OrderedDict([('a', Ellipsis), ('c', Ellipsis)]), [0, 2])

    Now only the arguments in ``argnum`` are considered, in particular the ``argnum`` input
    is simply sorted. In ``encoding_args``, all argument names are paired with an ``Ellipsis``.
    If we skip both inputs, all QNode arguments are extracted:

    >>> _process_ids(None, None, circuit)
    (OrderedDict([('a', Ellipsis), ('b', Ellipsis), ('c', Ellipsis)]), [0, 1, 2])

    Note that ``x`` does not appear here, because it has a default value defined and thus is
    considered a keyword argument.

    """
    sig_pars = signature(qnode.func).parameters
    arg_names = list(sig_pars.keys())
    arg_names_no_def = [name for name, par in sig_pars.items() if par.default is par.empty]

    if encoding_args is None:
        if argnum is None:
            encoding_args = OrderedDict((name, ...) for name in arg_names_no_def)
            argnum = list(range(len(arg_names_no_def)))
        elif np.isscalar(argnum):
            encoding_args = OrderedDict({arg_names[argnum]: ...})
            argnum = [argnum]
        else:
            argnum = sorted(argnum)
            encoding_args = OrderedDict((arg_names[num], ...) for num in argnum)
    else:
        requested_names = set(encoding_args)
        if not all(name in arg_names for name in requested_names):
            raise ValueError(
                f"Not all names in {requested_names} are known. Known arguments: {arg_names}"
            )
        # Selection of requested argument names from sorted names
        if isinstance(encoding_args, set):
            encoding_args = OrderedDict(
                (name, ...) for name in arg_names if name in requested_names
            )
        else:
            encoding_args = OrderedDict(
                (name, encoding_args[name]) for name in arg_names if name in requested_names
            )
        argnum = [arg_names.index(name) for name in encoding_args]

    return encoding_args, argnum


def qnode_spectrum(qnode, encoding_args=None, argnum=None, decimals=8, validation_kwargs=None):
    r"""Compute the frequency spectrum of the Fourier representation of quantum circuits,
    including classical preprocessing.

    The circuit must only use gates as input-encoding gates that can be decomposed
    into single-parameter gates of the form :math:`e^{-i x_j G}` , which allows the
    computation of the spectrum by inspecting the gates' generators :math:`G`.
    The most important example of such single-parameter gates are Pauli rotations.

    The argument ``argnum`` controls which QNode arguments are considered as encoded
    inputs and the spectrum is computed only for these arguments.
    The input-encoding *gates* are those that are controlled by input-encoding QNode arguments.
    If no ``argnum`` is given, all QNode arguments are considered to be input-encoding
    arguments.

    .. note::

        Arguments of the QNode or parameters within an array-valued QNode argument
        that do not contribute to the Fourier series of the QNode
        with any frequency are considered as contributing with a constant term.
        That is, a parameter that does not control any gate has the spectrum ``[0]``.

    Args:
        qnode (pennylane.QNode): :class:`~.pennylane.QNode` to compute the spectrum for
        encoding_args (dict[str, list[tuple]], set): Parameter index dictionary;
            keys are argument names, values are index tuples for that argument
            or an ``Ellipsis``. If a ``set``, all values are set to ``Ellipsis``.
            The contained argument and parameter indices indicate the scalar variables
            for which the spectrum is computed
        argnum (list[int]): Numerical indices for arguments with respect to which
            to compute the spectrum
        decimals (int): number of decimals to which to round frequencies.
        validation_kwargs (dict): Keyword arguments passed to
            :func:`~.pennylane.math.is_independent` when testing for linearity of
            classical preprocessing in the QNode.

    Returns:
        function: Function which accepts the same arguments as the QNode.
        When called, this function will return a dictionary of dictionaries
        containing the frequency spectra per QNode parameter.

    **Details**

    A circuit that returns an expectation value of a Hermitian observable which depends on
    :math:`N` scalar inputs :math:`x_j` can be interpreted as a function
    :math:`f: \mathbb{R}^N \rightarrow \mathbb{R}`.
    This function can always be expressed by a Fourier-type sum

    .. math::

        \sum \limits_{\omega_1\in \Omega_1} \dots \sum \limits_{\omega_N \in \Omega_N}
        c_{\omega_1,\dots, \omega_N} e^{-i x_1 \omega_1} \dots e^{-i x_N \omega_N}

    over the *frequency spectra* :math:`\Omega_j \subseteq \mathbb{R},`
    :math:`j=1,\dots,N`. Each spectrum has the property that
    :math:`0 \in \Omega_j`, and the spectrum is symmetric
    (i.e., for every :math:`\omega \in \Omega_j` we have that :math:`-\omega \in\Omega_j`).
    If all frequencies are integer-valued, the Fourier sum becomes a *Fourier series*.

    As shown in `Vidal and Theis (2019) <https://arxiv.org/abs/1901.11434>`_ and
    `Schuld, Sweke and Meyer (2020) <https://arxiv.org/abs/2008.08605>`_,
    if an input :math:`x_j, j = 1 \dots N`,
    only enters into single-parameter gates of the form :math:`e^{-i x_j G}`
    (where :math:`G` is a Hermitian generator),
    the frequency spectrum :math:`\Omega_j` is fully determined by the eigenvalues
    of the generators :math:`G`. In many situations, the spectra are limited
    to a few frequencies only, which in turn limits the function class that the circuit
    can express.

    The ``qnode_spectrum`` function computes all frequencies that will
    potentially appear in the sets :math:`\Omega_1` to :math:`\Omega_N`.

    .. note::

        The ``qnode_spectrum`` function also supports
        preprocessing of the QNode arguments before they are fed into the gates,
        as long as this processing is *linear*. In particular, constant
        prefactors for the encoding arguments are allowed.

    **Example**

    Consider the following example, which uses non-trainable inputs ``x``, ``y`` and ``z``
    as well as trainable parameters ``w`` as arguments to the QNode.

    .. code-block:: python

        n_qubits = 3
        dev = qml.device("default.qubit", wires=n_qubits)

        @qml.qnode(dev)
        def circuit(x, y, z, w):
            for i in range(n_qubits):
                qml.RX(0.5*x[i], wires=i)
                qml.Rot(w[0,i,0], w[0,i,1], w[0,i,2], wires=i)
                qml.RY(2.3*y[i], wires=i)
                qml.Rot(w[1,i,0], w[1,i,1], w[1,i,2], wires=i)
                qml.RX(z, wires=i)
            return qml.expval(qml.PauliZ(wires=0))

    This circuit looks as follows:

    >>> x = np.array([1., 2., 3.])
    >>> y = np.array([0.1, 0.3, 0.5])
    >>> z = -1.8
    >>> w = np.random.random((2, n_qubits, 3))
    >>> print(qml.draw(circuit)(x, y, z, w))
    0: ──RX(0.5)──Rot(0.598, 0.949, 0.346)───RY(0.23)──Rot(0.693, 0.0738, 0.246)──RX(-1.8)──┤ ⟨Z⟩
    1: ──RX(1)────Rot(0.0711, 0.701, 0.445)──RY(0.69)──Rot(0.32, 0.0482, 0.437)───RX(-1.8)──┤
    2: ──RX(1.5)──Rot(0.401, 0.0795, 0.731)──RY(1.15)──Rot(0.756, 0.38, 0.38)─────RX(-1.8)──┤

    Applying the ``qnode_spectrum`` function to the circuit for
    the non-trainable parameters, we obtain:

    >>> res = qml.fourier.qnode_spectrum(circuit, argnum=[0, 1, 2])(x, y, z, w)
    >>> for inp, freqs in res.items():
    ...     print(f"{inp}: {freqs}")
    "x": {(0,): [-0.5, 0.0, 0.5], (1,): [-0.5, 0.0, 0.5], (2,): [-0.5, 0.0, 0.5]}
    "y": {(0,): [-2.3, 0.0, 2.3], (1,): [-2.3, 0.0, 2.3], (2,): [-2.3, 0.0, 2.3]}
    "z": {(): [-3.0, -2.0, -1.0, 0.0, 1.0, 2.0, 3.0]}

    .. note::
        While the Fourier spectrum usually does not depend
        on trainable circuit parameters or the actual values of the inputs,
        it may still change based on inputs to the QNode that alter the architecture
        of the circuit.

    .. UsageDetails::

        Above, we selected all input-encoding parameters for the spectrum computation, using
        the ``argnum`` keyword argument. We may also restrict the full analysis to a single
        QNode argument, again using ``argnum``:

        >>> res = qml.fourier.qnode_spectrum(circuit, argnum=[0])(x, y, z, w)
        >>> for inp, freqs in res.items():
        ...     print(f"{inp}: {freqs}")
        "x": {(0,): [-0.5, 0.0, 0.5], (1,): [-0.5, 0.0, 0.5], (2,): [-0.5, 0.0, 0.5]}

        Selecting arguments by name instead of index is possible via the
        ``encoding_args`` argument:

        >>> res = qml.fourier.qnode_spectrum(circuit, encoding_args={"y"})(x, y, z, w)
        >>> for inp, freqs in res.items():
        ...     print(f"{inp}: {freqs}")
        "y": {(0,): [-2.3, 0.0, 2.3], (1,): [-2.3, 0.0, 2.3], (2,): [-2.3, 0.0, 2.3]}

        Note that for array-valued arguments the spectrum for each element of the array
        is computed. A more fine-grained control is available by passing index tuples
        for the respective argument name in ``encoding_args``:

        >>> encoding_args = {"y": [(0,),(2,)]}
        >>> res = qml.fourier.qnode_spectrum(circuit, encoding_args=encoding_args)(x, y, z, w)
        >>> for inp, freqs in res.items():
        ...     print(f"{inp}: {freqs}")
        "y": {(0,): [-2.3, 0.0, 2.3], (2,): [-2.3, 0.0, 2.3]}

        .. warning::
            The ``qnode_spectrum`` function checks whether the classical preprocessing between
            QNode and gate arguments is linear by computing the Jacobian of the processing
            and applying :func:`~.pennylane.math.is_independent`. This makes it unlikely
            -- *but not impossible* -- that non-linear functions go undetected.
            The number of additional points at which the Jacobian is computed in the numerical
            test of ``is_independent`` as well as other options for this function
            can be controlled via ``validation_kwargs``.
            Furthermore, the QNode arguments *not* marked in ``argnum`` will not be
            considered in this test and if they resemble encoded inputs, the entire
            spectrum might be incorrect or the circuit might not even admit one.

        The ``qnode_spectrum`` function works in all interfaces:

        .. code-block:: python

            import tensorflow as tf

            dev = qml.device("default.qubit", wires=1)

            @qml.qnode(dev, interface='tf')
            def circuit(x):
                qml.RX(0.4*x[0], wires=0)
                qml.PhaseShift(x[1]*np.pi, wires=0)
                return qml.expval(qml.PauliZ(wires=0))

            x = tf.Variable([1., 2.])
            res = qml.fourier.qnode_spectrum(circuit)(x)

        >>> print(res)
        {"x": {(0,): [-0.4, 0.0, 0.4], (1,): [-3.14159, 0.0, 3.14159]}}

        Finally, compare ``qnode_spectrum`` with :func:`~.circuit_spectrum`, using
        the following circuit.

        .. code-block:: python

            dev = qml.device("default.qubit", wires=2)

            @qml.qnode(dev)
            def circuit(x, y, z):
                qml.RX(0.5*x**2, wires=0, id="x")
                qml.RY(2.3*y, wires=1, id="y0")
                qml.CNOT(wires=[1,0])
                qml.RY(z, wires=0, id="y1")
                return qml.expval(qml.PauliZ(wires=0))

        First, note that we assigned ``id`` labels to the gates for which we will use
        ``circuit_spectrum``. This allows us to choose these gates in the computation:

        >>> x, y, z = 0.1, 0.2, 0.3
        >>> circuit_spec_fn = qml.fourier.circuit_spectrum(circuit, encoding_gates=["x","y0","y1"])
        >>> circuit_spec = circuit_spec_fn(x, y, z)
        >>> for _id, spec in circuit_spec.items():
        ...     print(f"{_id}: {spec}")
        x: [-1.0, 0, 1.0]
        y0: [-1.0, 0, 1.0]
        y1: [-1.0, 0, 1.0]

        As we can see, the preprocessing in the QNode is not included in the simple spectrum.
        In contrast, the output of ``qnode_spectrum`` is:

        >>> adv_spec = qml.fourier.qnode_spectrum(circuit, encoding_args={"y", "z"})
        >>> for _id, spec in adv_spec.items():
        ...     print(f"{_id}: {spec}")
        y: {(): [-2.3, 0.0, 2.3]}
        z: {(): [-1.0, 0.0, 1.0]}

        Note that the values of the output are dictionaries instead of the spectrum lists, that
        they include the prefactors introduced by classical preprocessing, and
        that we would not be able to compute the advanced spectrum for ``x`` because it is
        preprocessed non-linearily in the gate ``qml.RX(0.5*x**2, wires=0, id="x")``.

    """
    # pylint: disable=too-many-branches,protected-access
    validation_kwargs = validation_kwargs or {}
    encoding_args, argnum = _process_ids(encoding_args, argnum, qnode)
    atol = 10 ** (-decimals) if decimals is not None else 1e-10
    # A map between Jacobian indices (contiguous) and arg names (may be discontiguous)
    arg_name_map = dict(enumerate(encoding_args))
    jac_fn = qml.transforms.classical_jacobian(
        qnode, argnum=argnum, expand_fn=qml.transforms.expand_multipar
    )

    @wraps(qnode)
    def wrapper(*args, **kwargs):
        # Compute classical Jacobian and assert preprocessing is linear
        if not qml.math.is_independent(jac_fn, qnode.interface, args, kwargs, **validation_kwargs):
            raise ValueError(
                "The Jacobian of the classical preprocessing in the provided QNode "
                "is not constant; only linear classical preprocessing is supported."
            )
        # After construction, check whether invalid operations (for a spectrum)
        # are present in the QNode
        for m in qnode.qtape.measurements:
            if m.return_type not in {qml.operation.Expectation, qml.operation.Probability}:
                raise ValueError(
                    f"The return_type {m.return_type.value} is not supported as it likely does "
                    "not admit a Fourier spectrum."
                )
        class_jacs = jac_fn(*args, **kwargs)
        spectra = {}
        tape = qml.transforms.expand_multipar(qnode.qtape)
        par_info = tape._par_info

        # Iterate over jacobians per argument
        for jac_idx, class_jac in enumerate(class_jacs):
            # Obtain argument name for the jacobian index
            arg_name = arg_name_map[jac_idx]
            # Extract requested parameter indices for the current argument
            if encoding_args[arg_name] is Ellipsis:
                # If no index for this argument is specified, request all parameters within
                # the argument (Recall () is a valid index for scalar-valued arguments here)
                requested_par_ids = set(product(*(range(sh) for sh in class_jac.shape[1:])))
            else:
                requested_par_ids = set(encoding_args[arg_name])
            # Each requested parameter at least "contributes" as a constant
            _spectra = {par_idx: {0} for par_idx in requested_par_ids}

            # Iterate over the axis of the current Jacobian that corresponds to the tape operations
            for op_idx, jac_of_op in enumerate(np.round(class_jac, decimals=decimals)):
                op = par_info[op_idx]["op"]
                # Find parameters that both were requested and feed into the operation
                if len(class_jac.shape) == 1:
                    # Scalar argument, only axis of Jacobian is for operations
                    if np.isclose(jac_of_op, 0.0, atol=atol, rtol=0):
                        continue
                    jac_of_op = {(): jac_of_op}
                    par_ids = {()}
                else:
                    # Array-valued argument
                    # Extract indices of parameters contributing to the current operation
                    par_ids = zip(*[map(int, _ids) for _ids in np.where(jac_of_op)])
                    # Exclude contributing parameters that were not requested
                    par_ids = set(par_ids).intersection(requested_par_ids)
                    if len(par_ids) == 0:
                        continue

                # Multi-parameter gates are not supported (we expanded the tape already)
                if len(op.parameters) != 1:
                    raise ValueError(
                        "Can only consider one-parameter gates as data-encoding gates; "
                        f"got {op.name}."
                    )
                spec = get_spectrum(op, decimals=decimals)

                # For each contributing parameter, rescale the operation's spectrum
                # and add it to the spectrum for that parameter
                for par_idx in par_ids:
                    scale = float(qml.math.abs(jac_of_op[par_idx]))
                    scaled_spec = [scale * f for f in spec]
                    _spectra[par_idx] = join_spectra(_spectra[par_idx], scaled_spec)

            # Construct the sorted spectrum also containing negative frequencies
            for idx, spec in _spectra.items():
                spec = sorted(spec)
                _spectra[idx] = [-freq for freq in spec[:0:-1]] + spec
            spectra[arg_name] = _spectra

        return spectra

    return wrapper



# Copyright 2018-2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This subpackage provides the functionality to perform differentiable Hartree-Fock calculations.
"""
from .basis_data import STO3G, atomic_numbers
from .basis_set import BasisFunction, atom_basis_data, mol_basis_data
from .hamiltonian import (
    generate_electron_integrals,
    generate_fermionic_hamiltonian,
    generate_hamiltonian,
)
from .hartree_fock import generate_scf, hf_energy, nuclear_energy
from .integrals import (
    contracted_norm,
    electron_repulsion,
    expansion,
    gaussian_kinetic,
    gaussian_overlap,
    generate_attraction,
    generate_kinetic,
    generate_overlap,
    generate_repulsion,
    nuclear_attraction,
    primitive_norm,
)
from .matrices import (
    generate_attraction_matrix,
    generate_core_matrix,
    generate_kinetic_matrix,
    generate_overlap_matrix,
    generate_repulsion_tensor,
    molecular_density_matrix,
)
from .molecule import Molecule



# Copyright 2018-2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions for circuit optimization."""
# pylint: disable=too-many-return-statements
from pennylane.math import allclose, sin, cos, arccos, arctan2, zeros, cast_like, stack
from pennylane.wires import Wires


def find_next_gate(wires, op_list):
    """Given a list of operations, finds the next operation that acts on at least one of
    the same set of wires, if present.

    Args:
        wires (Wires): A set of wires acted on by a quantum operation.
        op_list (list[Operation]): A list of operations that are implemented after the
            operation that acts on ``wires``.

    Returns:
        int or None: The index, in ``op_list``, of the earliest gate that uses one or more
        of the same wires, or ``None`` if no such gate is present.
    """
    next_gate_idx = None

    for op_idx, op in enumerate(op_list):
        if len(Wires.shared_wires([wires, op.wires])) > 0:
            next_gate_idx = op_idx
            break

    return next_gate_idx


def _yzy_to_zyz(middle_yzy):
    r"""Converts a set of angles representing a sequence of rotations RY, RZ, RY into
    an equivalent sequence of the form RZ, RY, RZ.

    Any rotation in 3-dimensional space (or, equivalently, any single-qubit unitary)
    can be expressed as a sequence of rotations about 3 axes in 12 different ways.
    Typically, the arbitrary single-qubit rotation is expressed as RZ(a) RY(b) RZ(c),
    but there are some situations, e.g., composing two such rotations, where we need
    to convert between representations. This function converts the angles of a sequence

    .. math::

       RY(y_1) RZ(z) RY(y_2)

    into the form

    .. math::

       RZ(z_1) RY(y) RZ(z_2)

    This is accomplished by first converting the rotation to quaternion form, and then
    extracting the desired set of angles.

    Args:
        y1 (float): The angle of the first ``RY`` rotation.
        z (float): The angle of the inner ``RZ`` rotation.
        y2 (float): The angle of the second ``RY`` rotation.

    Returns:
        tuple[float, float, float]: A list of rotation angles in the ZYZ representation.
    """
    if allclose(stack(middle_yzy), cast_like(zeros(3), stack(middle_yzy))):
        return stack([0.0, 0.0, 0.0])

    y1, z, y2 = middle_yzy[0], middle_yzy[1], middle_yzy[2]

    # First, compute the quaternion representation
    # https://ntrs.nasa.gov/api/citations/19770024290/downloads/19770024290.pdf
    qw = cos(z / 2) * cos(0.5 * (y1 + y2))
    qx = sin(z / 2) * sin(0.5 * (y1 - y2))
    qy = cos(z / 2) * sin(0.5 * (y1 + y2))
    qz = sin(z / 2) * cos(0.5 * (y1 - y2))

    # Now convert from YZY Euler angles to ZYZ angles
    # Source: http://bediyap.com/programming/convert-quaternion-to-euler-rotations/
    z1_arg1 = 2 * (qy * qz - qw * qx)
    z1_arg2 = 2 * (qx * qz + qw * qy)
    z1 = arctan2(z1_arg1, z1_arg2)

    y = arccos(qw ** 2 - qx ** 2 - qy ** 2 + qz ** 2)

    z2_arg1 = 2 * (qy * qz + qw * qx)
    z2_arg2 = -2 * (qx * qz - qw * qy)
    z2 = arctan2(z2_arg1, z2_arg2)

    return stack([z1, y, z2])


def fuse_rot_angles(angles_1, angles_2):
    """Computed the set of rotation angles that is obtained when composing
    two ``qml.Rot`` operations.

    The ``qml.Rot`` operation represents the most general single-qubit operation.
    Two such operations can be fused into a new operation, however the angular dependence
    is non-trivial.

    Args:
        angles_1 (float): A set of three angles for the first ``qml.Rot`` operation.
        angles_2 (float): A set of three angles for the second ``qml.Rot`` operation.

    Returns:
        array[float]: A tuple of rotation angles for a single ``qml.Rot`` operation
        that implements the same operation as the two sets of input angles.
    """
    # Check for all-zero instances; if there are some, we can just return the sum.
    are_angles_1_zero = allclose(angles_1, zeros(3))
    are_angles_2_zero = allclose(angles_2, zeros(3))

    if are_angles_1_zero or are_angles_2_zero:
        return stack([angles_1[i] + angles_2[i] for i in range(3)])

    # RZ(a) RY(b) RZ(c) fused with RZ(d) RY(e) RZ(f)
    # first produces RZ(a) RY(b) RZ(c+d) RY(e) RZ(f)
    left_z = angles_1[0]
    middle_yzy = angles_1[1], angles_1[2] + angles_2[0], angles_2[1]
    right_z = angles_2[2]

    # There are a few other cases to consider where things can be 0 and
    # avoid having to use the quaternion conversion routine
    # If b = 0, then we have RZ(a + c + d) RY(e) RZ(f)
    if allclose(middle_yzy[0], 0.0):
        # Then if e is close to zero, return a single rotation RZ(a + c + d + f)
        if allclose(middle_yzy[2], 0.0):
            return [left_z + middle_yzy[1] + right_z, 0.0, 0.0]
        return stack([left_z + middle_yzy[1], middle_yzy[2], right_z])

    # If c + d is close to 0, then we have the case RZ(a) RY(b + e) RZ(f)
    if allclose(middle_yzy[1], 0.0):
        # If b + e is 0, we have RZ(a + f)
        if allclose(middle_yzy[0] + middle_yzy[2], 0.0):
            return [left_z + right_z, 0.0, 0.0]
        return stack([left_z, middle_yzy[0] + middle_yzy[2], right_z])

    # If e is close to 0, then we have the case RZ(a) RY(b) RZ(c + d + f)
    # The case where b is 0 actually already covered in the first loop,
    # so only one case here.
    if allclose(middle_yzy[2], 0.0):
        return stack([left_z, middle_yzy[0], middle_yzy[1] + right_z])

    # Otherwise, we need to turn the RY(b) RZ(c+d) RY(e) into something
    # of the form RZ(u) RY(v) RZ(w)
    u, v, w = _yzy_to_zyz(middle_yzy)

    # Then we can combine to create
    # RZ(a + u) RY(v) RZ(w + f)
    return stack([left_z + u, v, w + right_z])



# Copyright 2018-2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module contains utilities and auxiliary functions which are shared
across the PennyLane submodules.
"""
# pylint: disable=protected-access,too-many-branches
from collections.abc import Iterable
import functools
import inspect
import itertools
import numbers
from operator import matmul

import numpy as np
import scipy

import pennylane as qml


def decompose_hamiltonian(H, hide_identity=False):
    r"""Decomposes a Hermitian matrix into a linear combination of Pauli operators.

    Args:
        H (array[complex]): a Hermitian matrix of dimension :math:`2^n\times 2^n`
        hide_identity (bool): does not include the :class:`~.Identity` observable within
            the tensor products of the decomposition if ``True``

    Returns:
        tuple[list[float], list[~.Observable]]: a list of coefficients and a list
        of corresponding tensor products of Pauli observables that decompose the Hamiltonian.

    **Example:**

    We can use this function to compute the Pauli operator decomposition of an arbitrary Hermitian
    matrix:

    >>> A = np.array(
    ... [[-2, -2+1j, -2, -2], [-2-1j,  0,  0, -1], [-2,  0, -2, -1], [-2, -1, -1,  0]])
    >>> coeffs, obs_list = decompose_hamiltonian(A)
    >>> coeffs
    [-1.0, -1.5, -0.5, -1.0, -1.5, -1.0, -0.5, 1.0, -0.5, -0.5]

    We can use the output coefficients and tensor Pauli terms to construct a :class:`~.Hamiltonian`:

    >>> H = qml.Hamiltonian(coeffs, obs_list)
    >>> print(H)
    (-1.0) [I0 I1]
    + (-1.5) [X1]
    + (-0.5) [Y1]
    + (-1.0) [Z1]
    + (-1.5) [X0]
    + (-1.0) [X0 X1]
    + (-0.5) [X0 Z1]
    + (1.0) [Y0 Y1]
    + (-0.5) [Z0 X1]
    + (-0.5) [Z0 Y1]

    This Hamiltonian can then be used in defining VQE problems using :class:`~.ExpvalCost`.
    """
    n = int(np.log2(len(H)))
    N = 2 ** n

    if H.shape != (N, N):
        raise ValueError(
            "The Hamiltonian should have shape (2**n, 2**n), for any qubit number n>=1"
        )

    if not np.allclose(H, H.conj().T):
        raise ValueError("The Hamiltonian is not Hermitian")

    paulis = [qml.Identity, qml.PauliX, qml.PauliY, qml.PauliZ]
    obs = []
    coeffs = []

    for term in itertools.product(paulis, repeat=n):
        matrices = [i._matrix() for i in term]
        coeff = np.trace(functools.reduce(np.kron, matrices) @ H) / N
        coeff = np.real_if_close(coeff).item()

        if not np.allclose(coeff, 0):
            coeffs.append(coeff)

            if not all(t is qml.Identity for t in term) and hide_identity:
                obs.append(
                    functools.reduce(
                        matmul,
                        [t(i) for i, t in enumerate(term) if t is not qml.Identity],
                    )
                )
            else:
                obs.append(functools.reduce(matmul, [t(i) for i, t in enumerate(term)]))

    return coeffs, obs


def sparse_hamiltonian(H, wires=None):
    r"""Computes the sparse matrix representation a Hamiltonian in the computational basis.

    Args:
        H (~.Hamiltonian): Hamiltonian operator for which the matrix representation should be
         computed
        wires (Iterable): Wire labels that indicate the order of wires according to which the matrix
         is constructed. If not profided, ``H.wires`` is used.

    Returns:
        coo_matrix: a sparse matrix in scipy coordinate list (COO) format with dimension
        :math:`(2^n, 2^n)`, where :math:`n` is the number of wires

    **Example:**

    This function can be used by passing a `qml.Hamiltonian` object as:

    >>> coeffs = [1, -0.45]
    >>> obs = [qml.PauliZ(0) @ qml.PauliZ(1), qml.PauliY(0) @ qml.PauliZ(1)]
    >>> H = qml.Hamiltonian(coeffs, obs)
    >>> H_sparse = sparse_hamiltonian(H)
    >>> H_sparse
    <4x4 sparse matrix of type '<class 'numpy.complex128'>'
        with 2 stored elements in COOrdinate format>

    The resulting sparse matrix can be either used directly or transformed into a numpy array:

    >>> H_sparse.toarray()
    array([[ 1.+0.j  ,  0.+0.j  ,  0.+0.45j,  0.+0.j  ],
           [ 0.+0.j  , -1.+0.j  ,  0.+0.j  ,  0.-0.45j],
           [ 0.-0.45j,  0.+0.j  , -1.+0.j  ,  0.+0.j  ],
           [ 0.+0.j  ,  0.+0.45j,  0.+0.j  ,  1.+0.j  ]])
    """
    if not isinstance(H, qml.Hamiltonian):
        raise TypeError("Passed Hamiltonian must be of type `qml.Hamiltonian`")

    if wires is None:
        wires = H.wires
    else:
        wires = qml.wires.Wires(wires)

    n = len(wires)
    matrix = scipy.sparse.coo_matrix((2 ** n, 2 ** n), dtype="complex128")

    coeffs = qml.math.toarray(H.data)

    for coeff, op in zip(coeffs, H.ops):

        obs = []
        for o in qml.operation.Tensor(op).obs:
            if len(o.wires) > 1:
                # todo: deal with operations created from multi-qubit operations such as Hermitian
                raise ValueError(
                    "Can only sparsify Hamiltonians whose constituent observables consist of "
                    "(tensor products of) single-qubit operators; got {}.".format(op)
                )
            obs.append(scipy.sparse.coo_matrix(o.matrix))

        mat = [scipy.sparse.eye(2, format="coo")] * n

        for i, wire in enumerate(op.wires):
            # find index of this wire in the ordering
            idx = wires.index(wire)
            mat[idx] = obs[i]

        matrix += functools.reduce(lambda i, j: scipy.sparse.kron(i, j, format="coo"), mat) * coeff

    return matrix.tocoo()


def _flatten(x):
    """Iterate recursively through an arbitrarily nested structure in depth-first order.

    See also :func:`_unflatten`.

    Args:
        x (array, Iterable, Any): each element of an array or an Iterable may itself be any of these types

    Yields:
        Any: elements of x in depth-first order
    """
    if isinstance(x, np.ndarray):
        yield from _flatten(x.flat)  # should we allow object arrays? or just "yield from x.flat"?
    elif isinstance(x, qml.wires.Wires):
        # Reursive calls to flatten `Wires` will cause infinite recursion (`Wires` atoms are `Wires`).
        # Since Wires are always flat, just yield.
        for item in x:
            yield item
    elif isinstance(x, Iterable) and not isinstance(x, (str, bytes)):
        for item in x:
            yield from _flatten(item)
    else:
        yield x


def _unflatten(flat, model):
    """Restores an arbitrary nested structure to a flattened iterable.

    See also :func:`_flatten`.

    Args:
        flat (array): 1D array of items
        model (array, Iterable, Number): model nested structure

    Raises:
        TypeError: if ``model`` contains an object of unsupported type

    Returns:
        Union[array, list, Any], array: first elements of flat arranged into the nested
        structure of model, unused elements of flat
    """
    if isinstance(model, (numbers.Number, str)):
        return flat[0], flat[1:]

    if isinstance(model, np.ndarray):
        idx = model.size
        res = np.array(flat)[:idx].reshape(model.shape)
        return res, flat[idx:]

    if isinstance(model, Iterable):
        res = []
        for x in model:
            val, flat = _unflatten(flat, x)
            res.append(val)
        return res, flat

    raise TypeError("Unsupported type in the model: {}".format(type(model)))


def unflatten(flat, model):
    """Wrapper for :func:`_unflatten`.

    Args:
        flat (array): 1D array of items
        model (array, Iterable, Number): model nested structure

    Raises:
        ValueError: if ``flat`` has more elements than ``model``
    """
    # pylint:disable=len-as-condition
    res, tail = _unflatten(np.asarray(flat), model)
    if len(tail) != 0:
        raise ValueError("Flattened iterable has more elements than the model.")
    return res


def _inv_dict(d):
    """Reverse a dictionary mapping.

    Returns multimap where the keys are the former values,
    and values are sets of the former keys.

    Args:
        d (dict[a->b]): mapping to reverse

    Returns:
        dict[b->set[a]]: reversed mapping
    """
    ret = {}
    for k, v in d.items():
        ret.setdefault(v, set()).add(k)
    return ret


def _get_default_args(func):
    """Get the default arguments of a function.

    Args:
        func (callable): a function

    Returns:
        dict[str, tuple]: mapping from argument name to (positional idx, default value)
    """
    signature = inspect.signature(func)
    return {
        k: (idx, v.default)
        for idx, (k, v) in enumerate(signature.parameters.items())
        if v.default is not inspect.Parameter.empty
    }


@functools.lru_cache()
def pauli_eigs(n):
    r"""Eigenvalues for :math:`A^{\otimes n}`, where :math:`A` is
    Pauli operator, or shares its eigenvalues.

    As an example if n==2, then the eigenvalues of a tensor product consisting
    of two matrices sharing the eigenvalues with Pauli matrices is returned.

    Args:
        n (int): the number of qubits the matrix acts on
    Returns:
        list: the eigenvalues of the specified observable
    """
    if n == 1:
        return np.array([1, -1])
    return np.concatenate([pauli_eigs(n - 1), -pauli_eigs(n - 1)])


def expand(matrix, original_wires, expanded_wires):
    r"""Expand a an operator matrix to more wires.

    Args:
        matrix (array): :math:`2^n \times 2^n` matrix where n = len(original_wires).
        original_wires (Sequence[int]): original wires of matrix
        expanded_wires (Union[Sequence[int], int]): expanded wires of matrix, can be shuffled.
            If a single int m is given, corresponds to list(range(m))

    Returns:
        array: :math:`2^m \times 2^m` matrix where m = len(expanded_wires).
    """
    if isinstance(expanded_wires, numbers.Integral):
        expanded_wires = list(range(expanded_wires))

    N = len(original_wires)
    M = len(expanded_wires)
    D = M - N

    if not set(expanded_wires).issuperset(original_wires):
        raise ValueError("Invalid target subsystems provided in 'original_wires' argument.")

    if qml.math.shape(matrix) != (2 ** N, 2 ** N):
        raise ValueError(
            "Matrix parameter must be of size (2**len(original_wires), 2**len(original_wires))"
        )

    dims = [2] * (2 * N)
    tensor = qml.math.reshape(matrix, dims)

    interface = qml.math.get_interface(tensor)

    if D > 0:
        extra_dims = [2] * (2 * D)
        identity = qml.math.reshape(qml.math.eye(2 ** D), extra_dims)

        if interface == "tensorflow":
            identity = qml.math.cast_like(identity, tensor)

        expanded_tensor = qml.math.tensordot(tensor, identity, axes=0)
        # Fix order of tensor factors
        expanded_tensor = qml.math.moveaxis(
            expanded_tensor, tuple(range(2 * N, 2 * N + D)), tuple(range(N, N + D))
        )
    else:
        expanded_tensor = tensor

    wire_indices = []
    for wire in original_wires:
        wire_indices.append(expanded_wires.index(wire))

    wire_indices = np.array(wire_indices)

    # Order tensor factors according to wires
    original_indices = np.array(range(N))
    expanded_tensor = qml.math.moveaxis(
        expanded_tensor, tuple(original_indices), tuple(wire_indices)
    )
    expanded_tensor = qml.math.moveaxis(
        expanded_tensor, tuple(original_indices + M), tuple(wire_indices + M)
    )

    return qml.math.reshape(expanded_tensor, (2 ** M, 2 ** M))


def expand_vector(vector, original_wires, expanded_wires):
    r"""Expand a vector to more wires.

    Args:
        vector (array): :math:`2^n` vector where n = len(original_wires).
        original_wires (Sequence[int]): original wires of vector
        expanded_wires (Union[Sequence[int], int]): expanded wires of vector, can be shuffled
            If a single int m is given, corresponds to list(range(m))

    Returns:
        array: :math:`2^m` vector where m = len(expanded_wires).
    """
    if isinstance(expanded_wires, numbers.Integral):
        expanded_wires = list(range(expanded_wires))

    N = len(original_wires)
    M = len(expanded_wires)
    D = M - N

    if not set(expanded_wires).issuperset(original_wires):
        raise ValueError("Invalid target subsystems provided in 'original_wires' argument.")

    if qml.math.shape(vector) != (2 ** N,):
        raise ValueError("Vector parameter must be of length 2**len(original_wires)")

    dims = [2] * N
    tensor = qml.math.reshape(vector, dims)

    if D > 0:
        extra_dims = [2] * D
        ones = qml.math.ones(2 ** D).reshape(extra_dims)
        expanded_tensor = qml.math.tensordot(tensor, ones, axes=0)
    else:
        expanded_tensor = tensor

    wire_indices = []
    for wire in original_wires:
        wire_indices.append(expanded_wires.index(wire))

    wire_indices = np.array(wire_indices)

    # Order tensor factors according to wires
    original_indices = np.array(range(N))
    expanded_tensor = qml.math.moveaxis(
        expanded_tensor, tuple(original_indices), tuple(wire_indices)
    )

    return qml.math.reshape(expanded_tensor, 2 ** M)



# Copyright 2018-2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This submodule contains the template for QFT.
"""
# pylint:disable=abstract-method,arguments-differ,protected-access

import functools
import numpy as np

import pennylane as qml
from pennylane.operation import AnyWires, Operation


class QFT(Operation):
    r"""QFT(wires)
    Apply a quantum Fourier transform (QFT).

    For the :math:`N`-qubit computational basis state :math:`|m\rangle`, the QFT performs the
    transformation

    .. math::

        |m\rangle \rightarrow \frac{1}{\sqrt{2^{N}}}\sum_{n=0}^{2^{N} - 1}\omega_{N}^{mn} |n\rangle,

    where :math:`\omega_{N} = e^{\frac{2 \pi i}{2^{N}}}` is the :math:`2^{N}`-th root of unity.

    **Details:**

    * Number of wires: Any (the operation can act on any number of wires)
    * Number of parameters: 0
    * Gradient recipe: None

    Args:
        wires (int or Iterable[Number, str]]): the wire(s) the operation acts on

    **Example**

    The quantum Fourier transform is applied by specifying the corresponding wires:

    .. code-block::

        wires = 3

        dev = qml.device('default.qubit',wires=wires)

        @qml.qnode(dev)
        def circuit_qft(basis_state):
            qml.BasisState(basis_state, wires=range(wires))
            qml.templates.QFT(wires=range(wires))
            return qml.state()

        circuit_qft([1.0, 0.0, 0.0])
    """
    num_params = 0
    num_wires = AnyWires
    par_domain = None
    grad_method = None

    @property
    def matrix(self):
        # Redefine the property here to allow for a custom _matrix signature
        mat = self._matrix(len(self.wires))
        if self.inverse:
            mat = mat.conj()
        return mat

    @classmethod
    @functools.lru_cache()
    def _matrix(cls, num_wires):
        dimension = 2 ** num_wires

        mat = np.zeros((dimension, dimension), dtype=np.complex128)
        omega = np.exp(2 * np.pi * 1j / dimension)

        for m in range(dimension):
            for n in range(dimension):
                mat[m, n] = omega ** (m * n)

        return mat / np.sqrt(dimension)

    @staticmethod
    def decomposition(wires):
        num_wires = len(wires)
        shifts = [2 * np.pi * 2 ** -i for i in range(2, num_wires + 1)]

        decomp_ops = []
        for i, wire in enumerate(wires):
            decomp_ops.append(qml.Hadamard(wire))

            for shift, control_wire in zip(shifts[: len(shifts) - i], wires[i + 1 :]):
                op = qml.ControlledPhaseShift(shift, wires=[control_wire, wire])
                decomp_ops.append(op)

        first_half_wires = wires[: num_wires // 2]
        last_half_wires = wires[-(num_wires // 2) :]

        for wire1, wire2 in zip(first_half_wires, reversed(last_half_wires)):
            swap = qml.SWAP(wires=[wire1, wire2])
            decomp_ops.append(swap)

        return decomp_ops

    def adjoint(self):
        return QFT(wires=self.wires).inv()
